# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""Unembedding samples with broken chains."""

from collections.abc import Callable
from heapq import heapify, heappop

import numpy as np

import dimod

__all__ = ['broken_chains',
           'discard',
           'majority_vote',
           'weighted_random',
           'MinimizeEnergy',
           ]


def broken_chains(samples, chains):
    """Find the broken chains.

    Args:
        samples (array_like):
            Samples as a nS x nV array_like object where nS is the number of samples and nV is the
            number of variables. The values should all be 0/1 or -1/+1.

        chains (list[array_like]):
            List of chains of length nC where nC is the number of chains.
            Each chain should be an array_like collection of column indices in samples.

    Returns:
        :obj:`numpy.ndarray`: A nS x nC boolean array. If i, j is True, then chain j in sample i is
        broken.

    Examples:
        >>> samples = np.array([[-1, +1, -1, +1], [-1, -1, +1, +1]], dtype=np.int8)
        >>> chains = [[0, 1], [2, 3]]
        >>> dwave.embedding.broken_chains(samples, chains)
        array([[True, True],
               [ False,  False]])

        >>> samples = np.array([[-1, +1, -1, +1], [-1, -1, +1, +1]], dtype=np.int8)
        >>> chains = [[0, 2], [1, 3]]
        >>> dwave.embedding.broken_chains(samples, chains)
        array([[False, False],
               [ True,  True]])

    """
    samples, labels = dimod.as_samples(samples)

    if labels != range(len(labels)):
        relabel = {v: idx for idx, v in enumerate(labels)}
        chains = [[relabel[v] for v in chain] for chain in chains]
    else:
        chains = list(map(list, chains))  # because we use them for indexing

    num_samples, num_variables = samples.shape
    num_chains = len(chains)

    broken = np.zeros((num_samples, num_chains), dtype=bool, order='F')

    for cidx, chain in enumerate(chains):
        if isinstance(chain, set):
            chain = list(chain)
        chain = np.asarray(chain)

        if chain.ndim > 1:
            raise ValueError("chains should be 1D array_like objects")

        # chains of length 1, or 0 cannot be broken
        if len(chain) <= 1:
            continue

        all_ = (samples[:, chain] == 1).all(axis=1)
        any_ = (samples[:, chain] == 1).any(axis=1)
        broken[:, cidx] = np.bitwise_xor(all_, any_)

    return broken


def discard(samples, chains):
    """Discard broken chains.

    Args:
        samples (samples_like):
            A collection of samples. `samples_like` is an extension of NumPy's
            array_like. See :func:`dimod.as_samples`.

        chains (list[array_like]):
            List of chains, where each chain is an array_like collection of
            the variables in the same order as their represention in the given
            samples.

    Returns:
        tuple: A 2-tuple containing:

            :obj:`numpy.ndarray`: Unembedded samples as an array of dtype 'int8'.
            Broken chains are discarded.

            :obj:`numpy.ndarray`: Indicies of rows with unbroken chains.

    Examples:
        This example unembeds two samples that chains nodes 0 and 1 to represent
        a single source node. The first sample has an unbroken chain, the second
        a broken chain.

        >>> import dimod
        >>> import numpy as np
        ...
        >>> chains = [(0, 1), (2,)]
        >>> samples = np.array([[1, 1, 0], [1, 0, 0]], dtype=np.int8)
        >>> unembedded, idx = dwave.embedding.discard(samples, chains)
        >>> unembedded
        array([[1, 0]], dtype=int8)
        >>> print(idx)
        [0]

    """
    samples, labels = dimod.as_samples(samples)

    if labels != range(len(labels)):
        relabel = {v: idx for idx, v in enumerate(labels)}
        chains = [[relabel[v] for v in chain] for chain in chains]
    else:
        chains = list(map(list, chains))  # because we use them for indexing

    num_samples, num_variables = samples.shape
    num_chains = len(chains)

    broken = broken_chains(samples, chains)

    unbroken_idxs, = np.where(~broken.any(axis=1))

    chain_variables = np.fromiter((np.asarray(tuple(chain))[0] if isinstance(chain, set) else np.asarray(chain)[0]
                                   for chain in chains),
                                  count=num_chains, dtype=int)

    return samples[np.ix_(unbroken_idxs, chain_variables)], unbroken_idxs


def majority_vote(samples, chains):
    """Unembed samples using the most common value for broken chains.

    Args:
        samples (samples_like):
            A collection of samples. `samples_like` is an extension of NumPy's
            array_like. See :func:`dimod.as_samples`.

        chains (list[array_like]):
            List of chains, where each chain is an array_like collection of
            the variables in the same order as their represention in the given
            samples.

    Returns:
        tuple: A 2-tuple containing:

            :obj:`numpy.ndarray`: Unembedded samples as an nS-by-nC array of
            dtype 'int8', where nC is the number of chains and nS the number
            of samples. Broken chains are resolved by setting the sample value to
            that of most the chain's elements or, for chains without a majority,
            an arbitrary value.

            :obj:`numpy.ndarray`: Indicies of the samples. Equivalent to
            :code:`np.arange(nS)` because all samples are kept and none added.

    Examples:
        This example unembeds samples from a target graph that chains nodes 0 and 1 to
        represent one source node and nodes 2, 3, and 4 to represent another.
        Both samples have one broken chain, with different majority values.

        >>> import dimod
        >>> import numpy as np
        ...
        >>> chains = [(0, 1), (2, 3, 4)]
        >>> samples = np.array([[1, 1, 0, 0, 1], [1, 1, 1, 0, 1]], dtype=np.int8)
        >>> unembedded, idx = dwave.embedding.majority_vote(samples, chains)
        >>> print(unembedded)
        [[1 0]
         [1 1]]
        >>> print(idx)
        [0 1]

    """
    samples, labels = dimod.as_samples(samples)

    if labels != range(len(labels)):
        relabel = {v: idx for idx, v in enumerate(labels)}
        chains = [[relabel[v] for v in chain] for chain in chains]
    else:
        chains = list(map(list, chains))  # because we use them for indexing

    num_samples, num_variables = samples.shape
    num_chains = len(chains)

    unembedded = np.empty((num_samples, num_chains), dtype='int8', order='F')

    # determine if spin or binary. If samples are all 1, then either method works, so we use spin
    # because it is faster
    if samples.all():  # spin-valued
        for cidx, chain in enumerate(chains):
            # we just need the sign for spin. We don't use np.sign because in that can return 0
            # and fixing the 0s is slow.
            unembedded[:, cidx] = 2*(samples[:, chain].sum(axis=1) >= 0) - 1
    else:  # binary-valued
        for cidx, chain in enumerate(chains):
            mid = len(chain) / 2
            unembedded[:, cidx] = (samples[:, chain].sum(axis=1) >= mid)

    return unembedded, np.arange(num_samples)  # we keep all of the samples in this case


def weighted_random(samples, chains):
    """Unembed samples using weighed random choice for broken chains.

    Args:
        samples (samples_like):
            A collection of samples. `samples_like` is an extension of NumPy's
            array_like. See :func:`dimod.as_samples`.

        chains (list[array_like]):
            List of chains, where each chain is an array_like collection of
            the variables in the same order as their represention in the given
            samples.

    Returns:
        tuple: A 2-tuple containing:

            :obj:`numpy.ndarray`: Unembedded samples as an nS-by-nC array of
            dtype 'int8', where nC is the number of chains and nS the number
            of samples. Broken chains are resolved by setting the sample value to
            a random value weighted by frequency of the value in the chain.

            :obj:`numpy.ndarray`: Indicies of the samples. Equivalent to
            :code:`np.arange(nS)` because all samples are kept
            and no samples are added.

    Examples:
        This example unembeds samples from a target graph that chains nodes 0 and 1 to
        represent one source node and nodes 2, 3, and 4 to represent another.
        The sample has broken chains for both source nodes.

        >>> import dimod
        >>> import numpy as np
        ...
        >>> chains = [(0, 1), (2, 3, 4)]
        >>> samples = np.array([[1, 0, 1, 0, 1]], dtype=np.int8)
        >>> unembedded, idx = dwave.embedding.weighted_random(samples, chains)  # doctest: +SKIP
        >>> unembedded  # doctest: +SKIP
        array([[1, 1]], dtype=int8)
        >>> idx  # doctest: +SKIP
        array([0, 1])

    """
    samples, labels = dimod.as_samples(samples)

    if labels != range(len(labels)):
        relabel = {v: idx for idx, v in enumerate(labels)}
        chains = [[relabel[v] for v in chain] for chain in chains]
    else:
        chains = list(map(list, chains))  # because we use them for indexing

    # it sufficies to choose a random index from each chain and use that to construct the matrix
    idx = [np.random.choice(chain) for chain in chains]

    num_samples, num_variables = samples.shape
    return samples[:, idx], np.arange(num_samples)  # we keep all of the samples in this case


class MinimizeEnergy(Callable):
    """Unembed samples by minimizing local energy for broken chains.

    Args:
        bqm (:class:`~dimod.BinaryQuadraticModel`).
            Binary quadratic model associated with the source graph.

        embedding (dict):
            Mapping from source graph to target graph as a dict of form {s: [t, ...], ...},
            where s is a source-model variable and t is a target-model variable.

    Examples:
        This example embeds from a triangular graph to a square graph,
        chaining target-nodes 2 and 3 to represent source-node c, and unembeds minimizing the
        energy for the samples. The first two sample have unbroken chains, the second two have
        broken chains.

        >>> import dimod
        >>> import numpy as np
        ...
        >>> h = {'a': 0, 'b': 0, 'c': 0}
        >>> J = {('a', 'b'): 1, ('b', 'c'): 1, ('a', 'c'): 1}
        >>> bqm = dimod.BinaryQuadraticModel.from_ising(h, J)
        >>> embedding = {'a': [0], 'b': [1], 'c': [2, 3]}
        >>> cbm = dwave.embedding.MinimizeEnergy(bqm, embedding)
        >>> samples = np.array([[+1, -1, +1, +1],
        ...                     [-1, -1, -1, -1],
        ...                     [-1, -1, +1, -1],
        ...                     [+1, +1, -1, +1]], dtype=np.int8)
        >>> chains = [embedding['a'], embedding['b'], embedding['c']]
        >>> unembedded, idx = cbm(samples, chains)
        >>> unembedded
        array([[ 1, -1,  1],
               [-1, -1, -1],
               [-1, -1,  1],
               [ 1,  1, -1]], dtype=int8)
        >>> idx
        array([0, 1, 2, 3])

    """
    def __init__(self, bqm, embedding):
        # this is an awkward construction but we need it to maintain consistency with the other
        # chain break methods
        self.chain_to_var = {frozenset(chain): v for v, chain in embedding.items()}

        self.bqm = bqm

    def __call__(self, samples, chains):
        """
        Args:
            samples (samples_like):
                A collection of samples. `samples_like` is an extension of NumPy's
                array_like. See :func:`dimod.as_samples`.

            chains (list[array_like]):
                List of chains, where each chain is an array_like collection of
                the variables in the same order as their represention in the given
                samples.

        Returns:
            tuple: A 2-tuple containing:

                :obj:`numpy.ndarray`: Unembedded samples as an nS-by-nC array of
                dtype 'int8', where nC is the number of chains and nS the number
                of samples. Broken chains are resolved by greedy energy descent.

                :obj:`numpy.ndarray`: Indicies of the samples. Equivalent to
                :code:`np.arange(nS)` because all samples are kept and none added.

        """
        samples, labels = dimod.as_samples(samples)

        if labels != range(len(labels)):
            relabel = {v: idx for idx, v in enumerate(labels)}
            chains = [[relabel[v] for v in chain] for chain in chains]
        else:
            chains = list(map(list, chains))  # because we use them for indexing

        chain_to_var = self.chain_to_var
        variables = [chain_to_var[frozenset(labels[c] for c in chain)]
                     for chain in chains]

        # we want the bqm by index
        bqm = self.bqm.relabel_variables({v: idx for idx, v in enumerate(variables)}, inplace=False)

        num_chains = len(chains)

        if bqm.vartype is dimod.SPIN:
            ZERO = -1
        else:
            ZERO = 0

        def _minenergy(arr):

            unbroken_arr = np.zeros((num_chains,), dtype=np.int8)
            broken = []

            for cidx, chain in enumerate(chains):
                eq1 = (arr[chain] == 1)
                if not np.bitwise_xor(eq1.all(), eq1.any()):
                    # not broken
                    unbroken_arr[cidx] = arr[chain][0]
                else:
                    broken.append(cidx)

            energies = []
            for cidx in broken:
                en = bqm.linear[cidx] + sum(unbroken_arr[idx] * bqm.adj[cidx][idx] for idx in bqm.adj[cidx])
                energies.append([-abs(en), en, cidx])
            heapify(energies)

            while energies:
                _, e, i = heappop(energies)

                unbroken_arr[i] = val = ZERO if e > 0 else 1

                for energy_triple in energies:
                    k = energy_triple[2]
                    if k in bqm.adj[i]:
                        energy_triple[1] += val * bqm.adj[i][k]
                    energy_triple[0] = -abs(energy_triple[1])

                heapify(energies)

            return unbroken_arr

        num_samples, num_variables = samples.shape
        return np.apply_along_axis(_minenergy, 1, samples), np.arange(num_samples)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import concurrent.futures
import warnings
from uuid import uuid4

import numpy as np
import dimod
import dwave.cloud.computation
from dwave.system import qpu_graph

try:
    #Simplest and efficient choice returning low energy states:
    from greedy import SteepestDescentSampler as SubstituteSampler
except ImportError:
    from dimod import SimulatedAnnealingSampler as SubstituteSampler


class MockDWaveSampler(dimod.Sampler, dimod.Structured):
    """Mock sampler modeled after DWaveSampler that can be used for tests.

    Properties and topology parameters are populated qualitatively matching
    online systems, and a placeholder sampler routine based on steepest descent
    is instantiated.

    Args:
        nodelist (iterable of ints, optional):
            List of nodes to include. When not specified, all nodes
            compatible with the default topology type and shape are
            added.

        edgelist (iterable of (int,int) tuples, optional):
            List of edges to include. When not specified, all edges
            compatible with the default topology type and shape are
            added.

        properties (dictionary, optional):
            A dictionary of properties, to update default values.

        broken_nodes (iterable of ints, optional):
            List of nodes to exclude (along with associated edges), for
            emulation of unyielded qubits. This parameter is made redundant
            by the use of nodelist.

        broken_edges (iterable of (int,int) tuples, optional):
            List of edges to exclude, for emulation of unyielded edges.
            This parameter is made redundant by the use of edgelist.

        topology_type (string, optional, default='chimera'):
            Supported options are 'chimera', 'pegasus' or 'zephyr'.
            The default is 'chimera' when the value is not specified as part 
            of the from the ``properties`` argument.

        topology_shape (string, optional):
            A list of three numbers [m,n,t] for Chimera, defaulted as [4,4,4]. 
            A list of one number [m] for Pegasus, defaulted as [3].
            A list of two numbers [m,t] for Zephyr, defaulted as [2,4].
            The default above apply only when the value is not 
            specified as part of the from the ``properties`` argument.

        parameter_warnings (bool, optional, default=True):
            The MockSampler is adaptive with respect to ``num_reads``,
            ``answer_mode`` and ``max_answers`` and ``label`` 
            parameters. By default ``initial_state`` can also be mocked, if
            dwave-greedy is installed. All other parameters are ignored and a 
            warning will be raised by default.

        exact_solver_cutoff (int, optional, default=:attr:`EXACT_SOLVER_CUTOFF_DEFAULT`):
            For problems smaller or equal in size to ``exact_solver_cutoff``, the
            first sample in any sampleset returned by the sampling routines
            is replaced by a reproducible ground state (determined exactly with
            a brute-force :class:`~dimod.ExactSolver`). Only small cut offs
            should be used since solution time increases exponentially with
            problem size.
            Set ``exact_solver_cutoff`` to zero to disable exact ground state
            calculation.

    Examples
        The first example creates a MockSampler without reference to a
        particular online system, with a fully-yielded Chimera C16 structure.
        The second example creates a MockSampler from the default DWaveSampler:
        nodelist, edgelist and properties are derived from the sampler.
        In each case a single qubit problem is sampled.

        >>> from dwave.system.testing import MockDWaveSampler
        >>> mock_sampler = MockDWaveSampler(topology_type='chimera',
        ...                                 topology_shape=[16,16,4])
        >>> ss = mock_sampler.sample_ising({mock_sampler.nodelist[0] : -1}, {},
        ...                                num_reads = 100)
        >>> print(ss.first.energy)
        -1
        ...
        >>> from dwave.system import DWaveSampler
        >>> sampler = DWaveSampler()
        >>> mock_sampler = MockDWaveSampler.from_qpu_sampler(sampler)
        >>> ss = mock_sampler.sample_ising({mock_sampler.nodelist[0] : -1}, {},
        ...                                num_reads = 100)
        >>> print(ss.first.energy)
        -1

    """
    # Feature suggestion - add seed as an optional input, to allow reproducibility.

    nodelist = None
    edgelist = None
    properties = None
    parameters = None

    # by default, use ExactSolver for problems up to size (inclusive):
    EXACT_SOLVER_CUTOFF_DEFAULT = 16

    def __init__(self,
                 nodelist=None, edgelist=None, properties=None,
                 broken_nodes=None, broken_edges=None,
                 topology_type=None, topology_shape=None,
                 parameter_warnings=True,
                 exact_solver_cutoff=EXACT_SOLVER_CUTOFF_DEFAULT,
                 **config):
        self.parameter_warnings = parameter_warnings
        self.exact_solver_cutoff = exact_solver_cutoff

        #Parse or default topology dependent arguments:
        if properties is not None and 'topology' in properties:
            if ('type' not in properties['topology']
                or 'shape' not in properties['topology']):
                raise ValueError("'shape' and 'type' should be keys in "
                                 "properties['topology']")
            if topology_type is not None and topology_type != properties['topology']['type']:
                raise ValueError("topology_type must be compatible with "
                                 "properties['topology']['type'] when specified")
            topology_type = properties['topology']['type']
            if topology_shape is not None and topology_shape != properties['topology']['shape']:
                raise ValueError("topology_shape must be compatible with " 
                                 "properties['topology']['shape'] when specified")
            topology_shape = properties['topology']['shape']
        else:
            if topology_type is None:
                topology_type = 'chimera'
            shape_defaults = {'chimera': [4,4,4],
                              'pegasus': [3],
                              'zephyr': [2,4]}
            if topology_type in shape_defaults:
                if topology_shape is None:
                    topology_shape = shape_defaults[topology_type]
            else:
                raise ValueError("Only 'chimera', 'pegasus' and 'zephyr' "
                                 "topologies are supported")
        self.properties = {
            'chip_id': 'MockDWaveSampler',
            'topology': {'type': topology_type, 'shape': topology_shape}
        }

        #Create graph object, introduce defects per input arguments
        if nodelist is not None:
            self.nodelist = nodelist.copy()
        if edgelist is not None:
            self.edgelist = edgelist.copy()
        # Note that self.to_networkx_graph would point to an inherited
        # version rather than the class method here, without topology
        # information, for clarity helper function is separated.
        solver_graph = qpu_graph(self.properties['topology']['type'],
                                 self.properties['topology']['shape'],
                                 self.nodelist, self.edgelist)

        if broken_nodes is None and broken_edges is None:
            self.nodelist = sorted(solver_graph.nodes)
            self.edgelist = sorted(tuple(sorted(edge))
                                   for edge in solver_graph.edges)
        else:
            if broken_nodes is None:
                broken_nodes = []
            self.nodelist = sorted(set(solver_graph.nodes).difference(broken_nodes))
            if broken_edges is None:
                broken_edges = []
            self.edgelist = sorted(tuple(sorted((u, v))) for
                                   u, v in solver_graph.edges
                                   if u not in broken_nodes
                                   and v not in broken_nodes
                                   and (u, v) not in broken_edges
                                   and (v, u) not in broken_edges)
        #Finalize yield-dependent properties:
        self.properties.update({
            'num_qubits': len(solver_graph),
            'qubits': self.nodelist.copy(),
            'couplers': self.edgelist.copy(),
            'anneal_offset_ranges': [[-0.5, 0.5] if i in self.nodelist
                                     else [0, 0] for i in range(len(self.nodelist))]})
        # Non-topology-dependent properties and parameters mocked from
        # Advantage_system4.1 accessed February 10th 2022
        # with simplified lists for large parameters, and modified
        # topology arguments per MockSolver initialization:
        # See also:
        # https://docs.dwavesys.com/docs/latest/c_solver_parameters.html

        self.parameters = {
            'anneal_offsets': ['parameters'],
            'anneal_schedule': ['parameters'],
            'annealing_time': ['parameters'],
            'answer_mode': ['parameters'],
            'auto_scale': ['parameters'],
            'flux_biases': ['parameters'],
            'flux_drift_compensation': ['parameters'],
            'h_gain_schedule': ['parameters'],
            'initial_state': ['parameters'],
            'max_answers': ['parameters'],
            'num_reads': ['parameters'],
            'num_spin_reversal_transforms': ['parameters'],
            'programming_thermalization': ['parameters'],
            'readout_thermalization': ['parameters'],
            'reduce_intersample_correlation': ['parameters'],
            'reinitialize_state': ['parameters'],
            'warnings': [],
            'label': [],
        }

        self.properties.update({
            'h_range': [-4.0, 4.0],
            'j_range': [-1.0, 1.0],
            'supported_problem_types': ['ising', 'qubo'],
            'parameters': {
                'anneal_offsets':
                'Anneal offsets for each working qubit, formatted as a list, '
                'with NaN specified for unused qubits.',
                'anneal_schedule':
                "Annealing schedule formatted as a piecewise linear list of "
                "floating-point pairs of 't' and 's'.",
                'annealing_time':
                'Quantum annealing duration, in microseconds, as a positive '
                'floating point number.',
                'answer_mode':
                "Format of returned answers, as 'histogram' or 'raw' samples.",
                'auto_scale':
                'Automatic rescaling of h and J values to their available '
                'range, as a boolean flag.',
                'flux_biases':
                'Flux biases for each working qubit, in normalized offset '
                'units, formatted as a list for all qubits.',
                'flux_drift_compensation':
                'Activation of flux drift compensation, as a boolean flag.',
                'h_gain_schedule':
                "h-gain schedule formatted as a piecewise linear list of "
                "floating-point pairs of 't' and 'g'.",
                'initial_state':
                'Initial states to use for a reverse-anneal request, as a list '
                'of qubit index and state.',
                'max_answers':
                'Maximum number of answers to return.',
                'num_reads':
                'Number of states to read (answers to return), as a positive '
                'integer.',
                'num_spin_reversal_transforms':
                'Number of spin-reversal transforms (gauge transformations) to '
                'perform.',
                'programming_thermalization':
                'Time in microseconds to wait after programming the processor '
                'in order for it to cool back to base temperature, as a '
                'positive floating point number.',
                'readout_thermalization':
                'Time in microseconds to wait after each state is read from '
                'the processor in order for it to cool back to base '
                'temperature, as a positive floating point number.',
                'reduce_intersample_correlation':
                'Addition of pauses between samples, as a boolean flag.',
                'reinitialize_state':
                'Reapplication of the initial_state for every read in reverse '
                'annealing, as a boolean flag.'},
            'vfyc': False,
            'anneal_offset_step': -0.0001500217998314891,
            'anneal_offset_step_phi0': 1.4303846404537006e-05,
            'annealing_time_range': [0.02, 83000.0],
            'default_annealing_time': 20.0,
            'default_programming_thermalization': 1000.0,
            'default_readout_thermalization': 0.0,
            'extended_j_range': [-2.0, 1.0],
            'h_gain_schedule_range': [-3.0, 3.0],
            'max_anneal_schedule_points': 50,
            'max_h_gain_schedule_points': 20,
            'num_reads_range': [1, 10000],
            'per_qubit_coupling_range': [-18.0, 15.0],
            'problem_run_duration_range': [0.0, 10000000.0],
            # TODO: populate using `dwave.cloud.testing.mocks.qpu_problem_timing_data`
            # (when the mock is merged and released)
            'problem_timing_data': None,
            'programming_thermalization_range': [0.0, 10000.0],
            'readout_thermalization_range': [0.0, 10000.0],
            'tags': [],
            'category': 'qpu',
            'quota_conversion_rate': 1,
        })

        if properties is not None:
            # provided properties overwrite defaults, note that
            # particular care should be taken with respect to
            # topology-dependent arguments:
            self.properties.update(properties)

    @classmethod
    def from_qpu_sampler(cls, sampler):
        return cls(properties=sampler.properties,
                   nodelist=sampler.nodelist,
                   edgelist=sampler.edgelist)

    @dimod.bqm_structured
    def sample(self, bqm, **kwargs):

        # Check kwargs compatibility with parameters and substitute sampler:
        mocked_parameters={'answer_mode',
                           'max_answers',
                           'num_reads',
                           'label'}
        if SubstituteSampler is not dimod.SimulatedAnnealingSampler:
            mocked_parameters.add('initial_state')
            # steepest greedy descent is the best substitute to
            # standardize upon. A few additional considerations as
            # dwave-greedy is modified and processor scale grows:
            # (1) For large sample sets, it would be more efficient to
            # handle ``answer_mode`` and ``max_answers`` within the
            # lower-level optimized code - just as they are handled server
            # side in QPU calls.
            # (2) We could also consider using 'large_sparse_opt' to
            # exploit fixed connectivity at large scale for current QPU
            # designs, but unlikely to be a significant inefficiency.

            # The fallback sampler (SA) is an inferior choice, but
            # greedy may not always be installed, unlike dimod.
        for kw in kwargs:
            if kw in self.parameters:
                if self.parameter_warnings and kw not in mocked_parameters:
                    warnings.warn(f'{kw!r} parameter is valid for DWaveSampler(), '
                                  'but not mocked in MockDWaveSampler().')
            else:
                raise ValueError(f'kwarg {kw!r} invalid for MockDWaveSampler()')

        # Timing values are for demonstration only. These could be made
        # adaptive to sampler parameters and mocked topology in principle.
        # we should do it in a follow-up PR, using estimate_qpu_access_time,
        # once dwavesystems/dwave-cloud-client#530 is merged.
        info = dict(problem_id=str(uuid4()),
                    timing={'qpu_sampling_time': 82.08,
                            'qpu_anneal_time_per_sample': 20.0,
                            'qpu_readout_time_per_sample': 41.54,
                            'qpu_access_time': 8550.28,
                            'qpu_access_overhead_time': 9340.72,
                            'qpu_programming_time': 8468.2,
                            'qpu_delay_time_per_sample': 20.54,
                            'total_post_processing_time': 1124.0,
                            'post_processing_overhead_time': 1124.0})
        label = kwargs.get('label')
        if label is not None:
            info.update(problem_label=label)

        #Special handling of flux_biases, for compatibility with virtual graphs

        flux_biases = kwargs.get('flux_biases')
        if flux_biases is not None:
            self.flux_biases_flag = True

        substitute_kwargs = {'num_reads' : kwargs.get('num_reads')}
        if substitute_kwargs['num_reads'] is None:
            substitute_kwargs['num_reads'] = 1

        if SubstituteSampler is not dimod.SimulatedAnnealingSampler:
            initial_state = kwargs.get('initial_state')
            if initial_state is not None:
                # Initial state format is a list of (qubit,values)
                # value=3 denotes an unused variable (should be absent
                # from bqm). 
                # Convert to format for substitute (NB: plural key)
                substitute_kwargs['initial_states'] = (
                    np.array([pair[1] for pair in initial_state
                              if pair[1]!=3],dtype=float),
                    [pair[0] for pair in initial_state if pair[1]!=3])

        ss = SubstituteSampler().sample(bqm, **substitute_kwargs)
        ss.info.update(info)

        # determine ground state exactly for small problems
        if 0 < len(bqm) <= self.exact_solver_cutoff and len(ss) >= 1:
            ground = dimod.ExactSolver().sample(bqm).truncate(1)
            ss.record[0].sample = ground.record[0].sample
            ss.record[0].energy = ground.record[0].energy

        answer_mode = kwargs.get('answer_mode')
        if answer_mode is None or answer_mode == 'histogram':
            # Default for DWaveSampler() is 'histogram'
            ss = ss.aggregate()

        max_answers = kwargs.get('max_answers')
        if max_answers is not None:
            # Truncate sampleset if requested. Do not reorder (per DWaveSampler())
            ss = ss.truncate(max_answers)

        return ss

    def to_networkx_graph(self):
        return qpu_graph(self.properties['topology']['type'],
                         self.properties['topology']['shape'],
                         self.nodelist, self.edgelist)
    
    
class MockLeapHybridDQMSampler:
    """Mock sampler modeled after LeapHybridDQMSampler that can be used for tests."""
    def __init__(self, **config):
        self.parameters = {'time_limit': ['parameters'],
                           'label': []}

        self.properties = {'category': 'hybrid',
                           'supported_problem_types': ['dqm'],
                           'quota_conversion_rate': 20,
                           'minimum_time_limit': [[20000, 5.0],
                                                  [100000, 6.0],
                                                  [200000, 13.0],
                                                  [500000, 34.0],
                                                  [1000000, 71.0],
                                                  [2000000, 152.0],
                                                  [5000000, 250.0],
                                                  [20000000, 400.0],
                                                  [250000000, 1200.0]],
                           'maximum_time_limit_hrs': 24.0,
                           'maximum_number_of_variables': 3000,
                           'maximum_number_of_biases': 3000000000}

    def sample_dqm(self, dqm, **kwargs):
        num_samples = 12    # min num of samples from dqm solver
        samples = np.empty((num_samples, dqm.num_variables()), dtype=int)

        for vi, v in enumerate(dqm.variables):
            samples[:, vi] = np.random.choice(dqm.num_cases(v), size=num_samples)

        return dimod.SampleSet.from_samples((samples, dqm.variables),
                                             vartype='DISCRETE',
                                             energy=dqm.energies(samples))

    def min_time_limit(self, dqm):
        # not caring about the problem, just returning the min
        return self.properties['minimum_time_limit'][0][1]


class MockLeapHybridSolver:

    properties = {'supported_problem_types': ['bqm'],
                  'minimum_time_limit': [[1, 1.0], [1024, 1.0],
                                         [4096, 10.0], [10000, 40.0]],
                  'parameters': {'time_limit': None},
                  'category': 'hybrid',
                  'quota_conversion_rate': 1}

    supported_problem_types = ['bqm']

    def upload_bqm(self, bqm, **parameters):
        bqm_adjarray = dimod.serialization.fileview.load(bqm)
        future = concurrent.futures.Future()
        future.set_result(bqm_adjarray)
        return future

    def sample_bqm(self, sapi_problem_id, time_limit):
        
        bqm = dimod.BQM(sapi_problem_id.linear,
                                    sapi_problem_id.quadratic,
                                    sapi_problem_id.offset,
                                    sapi_problem_id.vartype)
        sampler = SubstituteSampler()
        result = sampler.sample(bqm, timeout=1000*int(time_limit))
        result = sampler.sample(bqm, timeout=1000*int(time_limit))
        future = dwave.cloud.computation.Future('fake_solver', None)
        future._result = {'sampleset': result, 'problem_type': 'bqm'}
        return future



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import itertools
import unittest
import warnings

from collections.abc import Mapping
from unittest import mock

import dimod
import dwave_networkx as dnx
from parameterized import parameterized_class

import dwave.embedding

from dwave.system.composites import (EmbeddingComposite,
                                     FixedEmbeddingComposite,
                                     LazyFixedEmbeddingComposite,
                                     LazyEmbeddingComposite,
                                     AutoEmbeddingComposite,
                                     )

from dwave.system.testing import MockDWaveSampler
from dwave.embedding import chain_breaks
from dwave.system.warnings import ChainStrengthWarning

try:
    from dwave.preprocessing import ScaleComposite
except ImportError:
    # fall back on dimod of dwave.preprocessing is not installed
    from dimod import ScaleComposite


@dimod.testing.load_sampler_bqm_tests(EmbeddingComposite(MockDWaveSampler()))
class TestEmbeddingComposite(unittest.TestCase):
    def test_instantiation_smoketest(self):
        sampler = EmbeddingComposite(MockDWaveSampler())

        dimod.testing.assert_sampler_api(sampler)

    def test_sample_ising(self):
        sampler = EmbeddingComposite(MockDWaveSampler())

        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}

        response = sampler.sample_ising(h, J)

        # nothing failed and we got at least one response back
        self.assertGreaterEqual(len(response), 1)

        for sample in response.samples():
            self.assertIsInstance(sample, Mapping)
            self.assertEqual(set(sample), set(h))

        for sample, energy in response.data(['sample', 'energy']):
            self.assertIsInstance(sample, Mapping)
            self.assertEqual(set(sample), set(h))
            self.assertAlmostEqual(dimod.ising_energy(sample, h, J),
                                   energy)

    def test_sample_ising_unstructured_not_integer_labelled(self):
        sampler = EmbeddingComposite(MockDWaveSampler())

        h = {'a': -1., 'b': 2}
        J = {('a', 'b'): 1.5}

        response = sampler.sample_ising(h, J)

        # nothing failed and we got at least one response back
        self.assertGreaterEqual(len(response), 1)

        for sample in response.samples():
            for v in h:
                self.assertIn(v, sample)

        for sample, energy in response.data(['sample', 'energy']):
            self.assertAlmostEqual(dimod.ising_energy(sample, h, J),
                                   energy)

    def test_sample_qubo(self):
        sampler = EmbeddingComposite(MockDWaveSampler())

        Q = {(0, 0): .1, (0, 4): -.8, (4, 4): 1}

        response = sampler.sample_qubo(Q)

        # nothing failed and we got at least one response back
        self.assertGreaterEqual(len(response), 1)

        for sample in response.samples():
            for u, v in Q:
                self.assertIn(v, sample)
                self.assertIn(u, sample)

        for sample, energy in response.data(['sample', 'energy']):
            self.assertAlmostEqual(dimod.qubo_energy(sample, Q),
                                   energy)

    def test_max_cut(self):
        sampler = EmbeddingComposite(MockDWaveSampler())

        m = 2
        n = 2
        t = 2

        hoff = 2 * t
        voff = n * hoff
        mi = m * voff
        ni = n * hoff

        edges = []

        # tile edges
        edges.extend((k0, k1)
                     for i in range(0, ni, hoff)
                     for j in range(i, mi, voff)
                     for k0 in range(j, j + t)
                     for k1 in range(j + t, j + 2 * t))
        # horizontal edges
        edges.extend((k, k + hoff)
                     for i in range(t, 2 * t)
                     for j in range(i, ni - hoff, hoff)
                     for k in range(j, mi, voff))
        # vertical edges
        edges.extend((k, k + voff)
                     for i in range(t)
                     for j in range(i, ni, hoff)
                     for k in range(j, mi - voff, voff))

        J = {edge: 1 for edge in edges}
        h = {v: 0 for v in set().union(*J)}

        response = sampler.sample_ising(h, J)

    def test_singleton_variables(self):
        sampler = EmbeddingComposite(MockDWaveSampler())

        h = {0: -1., 4: 2}
        J = {}

        response = sampler.sample_ising(h, J)

        # nothing failed and we got at least one response back
        self.assertGreaterEqual(len(response), 1)

    def test_chain_break_method_customization(self):
        sampler = EmbeddingComposite(MockDWaveSampler())

        def mock_unembed(*args, **kwargs):
            self.assertIn('chain_break_method', kwargs)
            self.assertEqual(kwargs['chain_break_method'], chain_breaks.discard)
            mock_unembed.call_count += 1
            return dwave.embedding.unembed_sampleset(*args, **kwargs)

        mock_unembed.call_count = 0

        with mock.patch('dwave.system.composites.embedding.unembed_sampleset', mock_unembed):
            sampler.sample_ising({'a': 1}, {}, chain_break_method=chain_breaks.discard).resolve()

        self.assertEqual(mock_unembed.call_count, 1)

    def test_find_embedding_kwarg(self):
        child = dimod.StructureComposite(dimod.NullSampler(), [0, 1], [(0, 1)])

        def my_find_embedding(S, T):
            # does nothing
            return {v: [v] for v in set().union(*S)}

        sampler = EmbeddingComposite(child, find_embedding=my_find_embedding)

        # nothing breaks
        sampler.sample_ising({0: -1}, {})

    def test_embedding_parameters_construction(self):
        child = dimod.StructureComposite(dimod.NullSampler(), [0, 1], [(0, 1)])

        def my_find_embedding(S, T, a):
            assert a == -1
            return {v: [v] for v in set().union(*S)}

        sampler = EmbeddingComposite(child, find_embedding=my_find_embedding,
                                     embedding_parameters={'a': -1})

        # nothing breaks
        sampler.sample_ising({0: -1}, {})

    def test_embedding_parameters_sample(self):
        child = dimod.StructureComposite(dimod.NullSampler(), [0, 1], [(0, 1)])

        def my_find_embedding(S, T, a):
            assert a == -1
            return {v: [v] for v in set().union(*S)}

        sampler = EmbeddingComposite(child, find_embedding=my_find_embedding)

        # nothing breaks
        sampler.sample_ising({0: -1}, {}, embedding_parameters={'a': -1})

    def test_intermediate_composites(self):
        child = dimod.StructureComposite(dimod.NullSampler(), [0, 1], [(0, 1)])
        intermediate = dimod.TrackingComposite(child)
        sampler = EmbeddingComposite(intermediate)
        self.assertEqual(sampler.target_structure.nodelist, [0, 1])

    def test_scale_aware_scale_composite(self):

        nodelist = [0, 1, 2]
        edgelist = [(0, 1), (1, 2), (0, 2)]
        embedding = {'a': [0], 'b': [1, 2]}

        sampler = FixedEmbeddingComposite(
            dimod.TrackingComposite(
                ScaleComposite(
                    dimod.StructureComposite(
                        dimod.NullSampler(),
                        nodelist, edgelist))),
            embedding=embedding, scale_aware=True)

        _ = sampler.sample_ising({'a': 100, 'b': -100}, {'ab': 300})

        self.assertIn('ignored_interactions', sampler.child.input)
        ignored = sampler.child.input['ignored_interactions']

        self.assertTrue(ignored == [(1, 2)] or ignored == [(2, 1)])

    def test_return_embedding(self):
        nodelist = [0, 1, 2]
        edgelist = [(0, 1), (1, 2), (0, 2)]

        sampler = EmbeddingComposite(
            dimod.StructureComposite(dimod.NullSampler(), nodelist, edgelist))

        sampleset = sampler.sample_ising({'a': -1}, {'ac': 1}, return_embedding=True)

        self.assertIn('embedding', sampleset.info['embedding_context'])
        embedding = sampleset.info['embedding_context']['embedding']
        self.assertEqual(set(embedding), {'a', 'c'})

        self.assertIn('chain_break_method', sampleset.info['embedding_context'])
        self.assertEqual(sampleset.info['embedding_context']['chain_break_method'], 'majority_vote')  # the default

        self.assertIn('embedding_parameters', sampleset.info['embedding_context'])
        self.assertEqual(sampleset.info['embedding_context']['embedding_parameters'], {})  # the default

        self.assertIn('chain_strength', sampleset.info['embedding_context'])
        self.assertEqual(sampleset.info['embedding_context']['chain_strength'], 1.414)  # the default

        # default False
        sampleset = sampler.sample_ising({'a': -1}, {'ac': 1})
        self.assertNotIn('embedding_context', sampleset.info)

    def test_return_embedding_as_class_variable(self):
        nodelist = [0, 1, 2]
        edgelist = [(0, 1), (1, 2), (0, 2)]

        sampler = EmbeddingComposite(
            dimod.StructureComposite(dimod.NullSampler(), nodelist, edgelist))

        # temporarily change return_embedding_default
        EmbeddingComposite.return_embedding_default = True

        sampleset = sampler.sample_ising({'a': -1}, {'ac': 1})

        self.assertIn('embedding', sampleset.info['embedding_context'])
        embedding = sampleset.info['embedding_context']['embedding']
        self.assertEqual(set(embedding), {'a', 'c'})

        self.assertIn('chain_break_method', sampleset.info['embedding_context'])
        self.assertEqual(sampleset.info['embedding_context']['chain_break_method'], 'majority_vote')  # the default

        self.assertIn('embedding_parameters', sampleset.info['embedding_context'])
        self.assertEqual(sampleset.info['embedding_context']['embedding_parameters'], {})  # the default

        self.assertIn('chain_strength', sampleset.info['embedding_context'])
        self.assertEqual(sampleset.info['embedding_context']['chain_strength'], 1.414)  # the default

        # restore the default
        EmbeddingComposite.return_embedding_default = False

    def test_warnings(self):
        G = dnx.chimera_graph(12)

        sampler = EmbeddingComposite(
            dimod.StructureComposite(dimod.RandomSampler(), G.nodes, G.edges))

        # this will need chains lengths > 7
        J = {uv: -1 for uv in itertools.combinations(range(40), 2)}

        ss = sampler.sample_ising({}, J, warnings='SAVE')

        self.assertIn('warnings', ss.info)

    def test_warning_chain_strength(self):
        G = dnx.chimera_graph(12)

        sampler = EmbeddingComposite(
            dimod.StructureComposite(dimod.RandomSampler(), G.nodes, G.edges))

        # use a triangle so there is a chain of length 2
        J = {(0, 1): 100, (1, 2): 100, (0, 2): 1}

        ss = sampler.sample_ising({}, J, chain_strength=1, warnings='SAVE')
        self.assertIn('warnings', ss.info)

        count = 0
        for warning in ss.info['warnings']:
            if issubclass(warning['type'], ChainStrengthWarning):
                count += 1
        self.assertEqual(count, 1)

    def test_warnings_chain_strength_len1(self):
        G = dnx.chimera_graph(12)

        sampler = EmbeddingComposite(
            dimod.StructureComposite(dimod.RandomSampler(), G.nodes, G.edges))

        J = {(0, 1): 100}

        ss = sampler.sample_ising({}, J, warnings='SAVE')
        self.assertIn('warnings', ss.info)

        # should have no chain length
        count = 0
        for warning in ss.info['warnings']:
            if issubclass(warning['type'], ChainStrengthWarning):
                count += 1
        self.assertEqual(count, 0)

    def test_warnings_chain_strength_dict(self):
        sampler = EmbeddingComposite(MockDWaveSampler())

        linear = {'a': -1, 'b': -2, 'c': -3}
        quadratic = {('a', 'b'): 1, ('a', 'c'): -1, ('b', 'c'): 2}
        bqm = dimod.BQM(linear, quadratic, 0, dimod.SPIN)

        chain_strength = {'a': 10, 'b': 20, 'c': 1.5}   
        ss = sampler.sample(bqm, chain_strength=chain_strength, warnings='SAVE')

        self.assertIn('warnings', ss.info)
        self.assertEqual(len(ss.info['warnings']), 1)

        warning = ss.info['warnings'][0]
        self.assertEqual(warning['type'], ChainStrengthWarning)

        interactions = warning['data']['source_interactions']
        self.assertEqual(len(interactions), 1)
        self.assertCountEqual(interactions[0], ('b','c'))

    def test_warnings_as_class_variable(self):
        G = dnx.chimera_graph(12)

        sampler = EmbeddingComposite(
            dimod.StructureComposite(dimod.RandomSampler(), G.nodes, G.edges))

        # this will need chains lengths > 7
        J = {uv: -1 for uv in itertools.combinations(range(40), 2)}

        EmbeddingComposite.warnings_default = 'SAVE'

        ss = sampler.sample_ising({}, J)

        self.assertIn('warnings', ss.info)

        EmbeddingComposite.warnings_default = 'IGNORE'  # restore default


class TestFixedEmbeddingComposite(unittest.TestCase):
    def test_without_embedding_and_adjacency(self):
        self.assertRaises(TypeError, lambda: FixedEmbeddingComposite(MockDWaveSampler()))

    def test_instantiation_empty_embedding(self):
        sampler = FixedEmbeddingComposite(MockDWaveSampler(), {})

        dimod.testing.assert_sampler_api(sampler)  # checks adj consistent with nodelist/edgelist

        self.assertEqual(sampler.edgelist, [])

        self.assertTrue(hasattr(sampler, 'embedding'))
        self.assertIn('embedding', sampler.properties)

    def test_instantiation_empty_adjacency(self):
        with self.assertWarns(DeprecationWarning):
            sampler = FixedEmbeddingComposite(MockDWaveSampler(), source_adjacency={})

        dimod.testing.assert_sampler_api(sampler)  # checks for attributes needed in a sampler

        self.assertEqual(sampler.edgelist, [])

        self.assertTrue(hasattr(sampler, 'embedding'))
        self.assertIn('embedding', sampler.properties)

    def test_instantiation_triangle(self):
        embedding = {'a': (0, 4), 'b': (1, 5), 'c': (2, 6)}
        sampler = FixedEmbeddingComposite(MockDWaveSampler(), embedding)
        self.assertEqual(embedding, sampler.embedding)

        dimod.testing.assert_sampler_api(sampler)  # checks adj consistent with nodelist/edgelist

        self.assertEqual(sampler.nodelist, ['a', 'b', 'c'])
        self.assertEqual(sampler.edgelist, [('a', 'b'), ('a', 'c'), ('b', 'c')])

    def test_sample_bqm_triangle(self):
        sampler = FixedEmbeddingComposite(MockDWaveSampler(), {'a': [0, 4], 'b': [1, 5], 'c': [2, 6]})

        resp = sampler.sample_ising({'a': 1, 'b': 1, 'c': 0}, {})

        self.assertEqual(set(resp.variables), {'a', 'b', 'c'})

    def test_adjacency(self):
        square_adj = {1: [2, 3], 2: [1, 4], 3: [1, 4], 4: [2, 3]}
        with self.assertWarns(DeprecationWarning):
            sampler = FixedEmbeddingComposite(MockDWaveSampler(), source_adjacency=square_adj)

        self.assertTrue(hasattr(sampler, 'adjacency'))
        self.assertTrue(hasattr(sampler, 'embedding'))
        self.assertIn('embedding', sampler.properties)

        self.assertEqual(sampler.nodelist, [1, 2, 3, 4])
        self.assertEqual(sampler.edgelist, [(1, 2), (1, 3), (2, 4), (3, 4)])

    def test_chain_break_method_customization(self):
        sampler = FixedEmbeddingComposite(MockDWaveSampler(), {'a': [0]})

        def mock_unembed(*args, **kwargs):
            self.assertIn('chain_break_method', kwargs)
            self.assertEqual(kwargs['chain_break_method'], chain_breaks.discard)
            mock_unembed.call_count += 1
            return dwave.embedding.unembed_sampleset(*args, **kwargs)

        mock_unembed.call_count = 0

        with mock.patch('dwave.system.composites.embedding.unembed_sampleset', mock_unembed):
            sampler.sample_ising({'a': 1}, {}, chain_break_method=chain_breaks.discard).resolve()

        self.assertEqual(mock_unembed.call_count, 1)

    def test_keyer(self):
        C4 = dnx.chimera_graph(4)
        nodelist = sorted(C4.nodes)
        edgelist = sorted(sorted(edge) for edge in C4.edges)

        child = dimod.StructureComposite(dimod.NullSampler(), nodelist, edgelist)

        embedding = {0: [49, 53], 1: [52], 2: [50]}

        sampler = FixedEmbeddingComposite(child, embedding)

        with self.assertRaises(dwave.embedding.exceptions.MissingChainError):
            sampler.sample_qubo({(1, 4): 1})

        sampler.sample_qubo({(1, 2): 1}).record  # sample and resolve future
        sampler.sample_qubo({(1, 1): 1}).record  # sample and resolve future

    def test_minimize_energy_chain_break_method(self):
        # bug report https://github.com/dwavesystems/dwave-system/issues/206
        Q = {(0, 1): 1, (0, 2): 1, (0, 3): 1, (1, 2): 1, (1, 3): 1, (2, 3): 1,
             (0, 0): 1, (1, 1): 1, (2, 2): 1, (3, 3): 1}
        S = {(0, 1): 1, (0, 2): 1, (0, 3): 1, (1, 2): 1, (1, 3): 1, (2, 3): 1}
        bqm = dimod.BinaryQuadraticModel.from_qubo(Q)

        G = dnx.chimera_graph(4)
        sampler = dimod.StructureComposite(dimod.ExactSolver(), G.nodes, G.edges)
        embedding = {0: [55], 1: [48], 2: [50, 53], 3: [52, 51]}
        composite = FixedEmbeddingComposite(sampler, embedding)

        cbm = chain_breaks.MinimizeEnergy(bqm, embedding)
        composite.sample(bqm, chain_break_method=cbm).resolve()


@dimod.testing.load_sampler_bqm_tests(lambda: LazyFixedEmbeddingComposite(MockDWaveSampler()))
class TestLazyFixedEmbeddingComposite(unittest.TestCase):
    def test_sample_instantiation(self):
        # Check that graph related values have not been instantiated
        sampler = LazyFixedEmbeddingComposite(MockDWaveSampler())
        self.assertIsNone(sampler.embedding)
        self.assertIsNone(sampler.nodelist)
        self.assertIsNone(sampler.edgelist)
        self.assertIsNone(sampler.adjacency)

        # Set up an and_gate BQM and sample
        Q = {('a', 'a'): 0.0, ('c', 'c'): 6.0, ('b', 'b'): 0.0, ('b', 'a'): 2.0, ('c', 'a'): -4.0, ('c', 'b'): -4.0}
        sampler.sample_qubo(Q)

        # Check that values have been populated
        self.assertIsNotNone(sampler.embedding)
        self.assertEqual(sampler.nodelist, ['a', 'b', 'c'])
        self.assertEqual(sampler.edgelist, [('a', 'b'), ('a', 'c'), ('b', 'c')])
        self.assertEqual(sampler.adjacency, {'a': {'b', 'c'}, 'b': {'a', 'c'}, 'c': {'a', 'b'}})

    def test_same_embedding(self):
        sampler = LazyFixedEmbeddingComposite(MockDWaveSampler())

        # Set up Ising and sample
        h = {'a': 1, 'b': 1, 'c': 1}
        J = {('a', 'b'): 3, ('b', 'c'): -2, ('a', 'c'): 1}
        sampler.sample_ising(h, J)

        # Store embedding
        prev_embedding = sampler.embedding

        # Set up QUBO of an or_gate
        Q = {('a', 'a'): 2.0, ('c', 'c'): 2.0, ('b', 'b'): 2.0, ('b', 'a'): 2.0, ('c', 'a'): -4.0, ('c', 'b'): -4.0}
        sampler.sample_qubo(Q)

        # Check that the same embedding is used
        self.assertEqual(sampler.embedding, prev_embedding)

    def test_ising(self):
        h = {0: 11, 5: 2}
        J = {(0, 5): -8}
        sampler = LazyFixedEmbeddingComposite(MockDWaveSampler())
        response = sampler.sample_ising(h, J)

        # Check embedding
        self.assertIsNotNone(sampler.embedding)
        self.assertEqual(sampler.nodelist, [0, 5])
        self.assertEqual(sampler.edgelist, [(0, 5)])
        self.assertEqual(sampler.adjacency, {0: {5}, 5: {0}})

        # Check that at least one response was found
        self.assertGreaterEqual(len(response), 1)

    def test_qubo(self):
        Q = {(1, 1): 1, (2, 2): 2, (3, 3): 3, (1, 2): 4, (2, 3): 5, (1, 3): 6}
        sampler = LazyFixedEmbeddingComposite(MockDWaveSampler())
        response = sampler.sample_qubo(Q)

        # Check embedding
        self.assertIsNotNone(sampler.embedding)
        self.assertEqual(sampler.nodelist, [1, 2, 3])
        self.assertEqual(sampler.edgelist, [(1, 2), (1, 3), (2, 3)])
        self.assertEqual(sampler.adjacency, {1: {2, 3}, 2: {1, 3}, 3: {1, 2}})

        # Check that at least one response was found
        self.assertGreaterEqual(len(response), 1)

    def test_sparse_qubo(self):
        # There is no relationship between nodes 2 and 3
        Q = {(1, 1): 1, (2, 2): 2, (3, 3): 3, (1, 2): 4, (1, 3): 6}
        sampler = LazyFixedEmbeddingComposite(MockDWaveSampler())
        response = sampler.sample_qubo(Q)

        # Check embedding
        self.assertIsNotNone(sampler.embedding)
        self.assertEqual(sampler.nodelist, [1, 2, 3])
        self.assertEqual(sampler.edgelist, [(1, 2), (1, 3)])

        # Check that at least one response was found
        self.assertGreaterEqual(len(response), 1)

    def test_chain_break_method_customization(self):
        sampler = LazyFixedEmbeddingComposite(MockDWaveSampler())

        def mock_unembed(*args, **kwargs):
            self.assertIn('chain_break_method', kwargs)
            self.assertEqual(kwargs['chain_break_method'], chain_breaks.discard)
            mock_unembed.call_count += 1
            return dwave.embedding.unembed_sampleset(*args, **kwargs)

        mock_unembed.call_count = 0

        with mock.patch('dwave.system.composites.embedding.unembed_sampleset', mock_unembed):
            sampler.sample_ising({'a': 1}, {}, chain_break_method=chain_breaks.discard).resolve()

        self.assertEqual(mock_unembed.call_count, 1)


class TestLazyEmbeddingComposite(unittest.TestCase):
    def test_deprecation_raise(self):
        # Temporarily mutate warnings filter
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")  # Cause all warnings to always be triggered.
            LazyEmbeddingComposite(MockDWaveSampler())  # Trigger warning

            # Verify deprecation warning
            assert len(w) == 1
            assert issubclass(w[-1].category, DeprecationWarning)
            assert "renamed" in str(w[-1].message)

    def test_ising_sample(self):
        h = {'a': 1, 'b': -2}
        J = {('a', 'b'): -3}
        sampler = LazyFixedEmbeddingComposite(MockDWaveSampler())
        response = sampler.sample_ising(h, J)

        # Check that at least one response was found
        self.assertGreaterEqual(len(response), 1)


@dimod.testing.load_sampler_bqm_tests(AutoEmbeddingComposite(MockDWaveSampler()))
class TestAutoEmbeddingComposite(unittest.TestCase):
    def test_broken_find_embedding(self):
        nodelist = [0, 1, 2, 3]
        edgelist = [(0, 1), (1, 2), (2, 3), (0, 3)]
        child = dimod.StructureComposite(dimod.NullSampler(), nodelist, edgelist)

        def find_embedding(*args, **kwargs):
            raise NotImplementedError

        sampler = AutoEmbeddingComposite(child, find_embedding=find_embedding)

        sampler.sample_ising({}, {(0, 1): -1})

        with self.assertRaises(NotImplementedError):
            sampler.sample_ising({}, {('a', 0): -1})

    def test_smoke(self):
        nodelist = [0, 1, 2, 3]
        edgelist = [(0, 1), (1, 2), (2, 3), (0, 3)]
        child = dimod.StructureComposite(dimod.NullSampler(), nodelist, edgelist)

        sampler = AutoEmbeddingComposite(child)

        sampler.sample_ising({}, {(0, 1): -1})

        sampler.sample_ising({}, {('a', 0): -1})

    def test_unstructured(self):
        child = dimod.NullSampler()

        sampler = AutoEmbeddingComposite(child)

        sampler.sample_ising({}, {(0, 1): -1})

        sampler.sample_ising({}, {('a', 0): -1}, embedding_parameters={})


@parameterized_class([
   dict(embedding_composite_class=EmbeddingComposite),
   dict(embedding_composite_class=AutoEmbeddingComposite),
])
class TestProblemLabelPropagation(unittest.TestCase):

    def test_problem_labelling(self):
        sampler = self.embedding_composite_class(MockDWaveSampler())

        self.assertIn('label', sampler.parameters)

        ss = sampler.sample_ising({}, {(0, 1): -1})
        self.assertNotIn('problem_label', ss.info)

        label = 'problem label'
        ss = sampler.sample_ising({}, {(0, 1): -1}, label=label)
        self.assertIn('problem_label', ss.info)
        self.assertEqual(ss.info['problem_label'], label)



# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import dwave_networkx as dnx


def coupling_groups(hardware_graph):
    """Generate groups of couplers for which a limit on total coupling applies for each group.

    Args:
        hardware_graph (:class:`networkx.Graph`): The hardware graph of a QPU.  Note that only
            :term:`Zephyr` graphs have coupling groups.

    Yields:
        Lists of tuples, where each list is a group of couplers in ``hardware_graph``.
    """

    if hardware_graph.graph.get('family') != 'zephyr':
        return

    relabel = dnx.zephyr_coordinates(hardware_graph.graph['rows']).linear_to_zephyr

    for q in hardware_graph:
        groups = [], []
        U, W, K, J, Z = relabel(q)

        for p in hardware_graph[q]:
            u, w, k, j, z = relabel(p)
            if U != u:
                groups[2*Z+J+1-w].append((p, q))
            elif J != j:
                groups[Z+J-z].append((p, q))
            else:
                groups[(z-Z)//2].append((p, q))

        yield from groups



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from math import ceil, sqrt

from dwave_networkx import chimera_graph, draw_chimera

__all__ = ['draw_chimera_bqm']

def draw_chimera_bqm(bqm, width=None, height=None):
    """Draws a Chimera Graph representation of a Binary Quadratic Model.

    If cell width and height not provided assumes square cell dimensions.
    Throws an error if drawing onto a Chimera graph of the given dimensions fails.

    Args:
        bqm (:obj:`dimod.BinaryQuadraticModel`):
            Should be equivalent to a Chimera graph or a subgraph of a Chimera graph produced by dnx.chimera_graph.
            The nodes and edges should have integer variables as in the dnx.chimera_graph.
        width (int, optional):
            An integer representing the number of cells of the Chimera graph will be in width.
        height (int, optional):
            An integer representing the number of cells of the Chimera graph will be in height.

    Examples:
        >>> from dwave.embedding.drawing import draw_chimera_bqm
        >>> from dimod import BinaryQuadraticModel
        >>> Q={(0, 0): 2, (1, 1): 1, (2, 2): 0, (3, 3): -1, (4, 4): -2, (5, 5): -2, (6, 6): -2, (7, 7): -2,
        ... (0, 4): 2, (0, 4): -1, (1, 7): 1, (1, 5): 0, (2, 5): -2, (2, 6): -2, (3, 4): -2, (3, 7): -2}
        >>> draw_chimera_bqm(BinaryQuadraticModel.from_qubo(Q), width=1, height=1)

    """

    linear = bqm.linear.keys()
    quadratic = bqm.quadratic.keys()

    if width is None and height is None:
        # Create a graph large enough to fit the input networkx graph.
        graph_size = ceil(sqrt((max(linear) + 1) / 8.0))
        width = graph_size
        height = graph_size

    if not width or not height:
        raise Exception("Both dimensions must be defined, not just one.")

    # A background image of the same size is created to show the complete graph.
    G0 = chimera_graph(height, width, 4)
    G = chimera_graph(height, width, 4)


    # Check if input graph is chimera graph shaped, by making sure that no edges are invalid.
    # Invalid edges can also appear if the size of the chimera graph is incompatible with the input graph in cell dimensions.
    non_chimera_nodes = []
    non_chimera_edges = []
    for node in linear:
        if not node in G.nodes:
            non_chimera_nodes.append(node)
    for edge in quadratic:
        if not edge in G.edges:
            non_chimera_edges.append(edge)

    linear_set = set(linear)
    g_node_set = set(G.nodes)

    quadratic_set = set(map(frozenset, quadratic))
    g_edge_set = set(map(frozenset, G.edges))

    non_chimera_nodes = linear_set - g_node_set
    non_chimera_edges = quadratic_set - g_edge_set

    if non_chimera_nodes or non_chimera_edges:
        raise Exception("Input graph is not a chimera graph: Nodes: %s Edges: %s" % (non_chimera_nodes, non_chimera_edges))


    # Get lists of nodes and edges to remove from the complete graph to turn the complete graph into your graph.
    remove_nodes = list(g_node_set - linear_set)
    remove_edges = list(g_edge_set - quadratic_set)

    # Remove the nodes and edges from the graph.
    for edge in remove_edges:
        G.remove_edge(*edge)
    for node in remove_nodes:
        G.remove_node(node)

    node_size = 100
    # Draw the complete chimera graph as the background.
    draw_chimera(G0, node_size=node_size*0.5, node_color='black', edge_color='black')
    # Draw your graph over the complete graph to show the connectivity.
    draw_chimera(G, node_size=node_size, linear_biases=bqm.linear, quadratic_biases=bqm.quadratic,
                     width=3)
    return


# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import dimod
import numpy as np

from dwave.embedding.chain_breaks import broken_chains


__all__ = ['target_to_source',
           'chain_to_quadratic',
           'chain_break_frequency',
           'adjacency_to_edges']


def target_to_source(target_adjacency, embedding):
    """Derive the source adjacency from an embedding and target adjacency.

    Args:
        target_adjacency (dict/:class:`networkx.Graph`):
            A dict of the form {v: Nv, ...} where v is a node in the target graph and Nv is the
            neighbors of v as an iterable. This can also be a networkx graph.

        embedding (dict):
            A mapping from a source graph to a target graph.

    Returns:
        dict: The adjacency of the source graph.

    Raises:
        ValueError: If any node in the target_adjacency is assigned more
            than  one node in the source graph by embedding.

    Examples:

        >>> target_adjacency = {0: {1, 3}, 1: {0, 2}, 2: {1, 3}, 3: {0, 2}}  # a square graph
        >>> embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
        >>> source_adjacency = dwave.embedding.target_to_source(target_adjacency, embedding)
        >>> # triangle graph:
        >>> source_adjacency   # doctest: +SKIP
        {'a': {'b', 'c'}, 'b': {'a', 'c'}, 'c': {'a', 'b'}}

        This function also works with networkx graphs.

        >>> import networkx as nx
        >>> target_graph = nx.complete_graph(5)
        >>> embedding = {'a': {0, 1, 2}, 'b': {3, 4}}
        >>> dwave.embedding.target_to_source(target_graph, embedding)   # doctest: +SKIP
        {'a': {'b'}, 'b': {'a'}}

    """
    # the nodes in the source adjacency are just the keys of the embedding
    source_adjacency = {v: set() for v in embedding}

    # we need the mapping from each node in the target to its source node
    reverse_embedding = {}
    for v, chain in embedding.items():
        for u in chain:
            if u in reverse_embedding:
                raise ValueError("target node {} assigned to more than one source node".format(u))
            reverse_embedding[u] = v

    # v is node in target, n node in source
    for v, n in reverse_embedding.items():
        neighbors = target_adjacency[v]

        # u is node in target
        for u in neighbors:

            # some nodes might not be assigned to chains
            if u not in reverse_embedding:
                continue

            # m is node in source
            m = reverse_embedding[u]

            if m == n:
                continue

            source_adjacency[n].add(m)
            source_adjacency[m].add(n)

    return source_adjacency


def chain_to_quadratic(chain, target_adjacency, chain_strength):
    """Determine the quadratic biases that induce the given chain.

    Args:
        chain (iterable):
            The variables that make up a chain.

        target_adjacency (dict/:class:`networkx.Graph`):
            Should be a dict of the form {s: Ns, ...} where s is a variable
            in the target graph and Ns is the set of neighbours of s.

        chain_strength (float):
            The magnitude of the quadratic bias that should be used to create chains.

    Returns:
        dict[edge, float]: The quadratic biases that induce the given chain.

    Raises:
        ValueError: If the variables in chain do not form a connected subgraph of target.

    Examples:

        >>> chain = {1, 2}
        >>> target_adjacency = {0: {1, 2}, 1: {0, 2}, 2: {0, 1}}
        >>> dimod.embedding.chain_to_quadratic(chain, target_adjacency, 1)
        {(1, 2): -1}

    """
    quadratic = {}  # we will be adding the edges that make the chain here

    # do a breadth first search
    seen = set()
    try:
        next_level = {next(iter(chain))}
    except StopIteration:
        raise ValueError("chain must have at least one variable")
    while next_level:
        this_level = next_level
        next_level = set()
        for v in this_level:
            if v not in seen:
                seen.add(v)

                for u in target_adjacency[v]:
                    if u not in chain:
                        continue
                    next_level.add(u)
                    if u != v and (u, v) not in quadratic:
                        quadratic[(v, u)] = -chain_strength

    if len(chain) != len(seen):
        raise ValueError('{} is not a connected chain'.format(chain))

    return quadratic


def chain_break_frequency(samples_like, embedding):
    """Determine the frequency of chain breaks in the given samples.

    Args:
        samples_like (samples_like/:obj:`dimod.SampleSet`):
            A collection of raw samples. 'samples_like' is an extension of NumPy's array_like.
            See :func:`dimod.as_samples`.

        embedding (dict):
            Mapping from source graph to target graph as a dict of form {s: {t, ...}, ...},
            where s is a source-model variable and t is a target-model variable.

    Returns:
        dict: Frequency of chain breaks as a dict in the form {s: f, ...},  where s
        is a variable in the source graph and float f the fraction
        of broken chains.

    Examples:
        This example embeds a single source node, 'a', as a chain of two target nodes (0, 1)
        and uses :func:`.chain_break_frequency` to show that out of two synthetic samples,
        one ([-1, +1]) represents a broken chain.

        >>> import numpy as np
        ...
        >>> samples = np.array([[-1, +1], [+1, +1]])
        >>> embedding = {'a': {0, 1}}
        >>> print(dwave.embedding.chain_break_frequency(samples, embedding)['a'])
        0.5


    """
    if isinstance(samples_like, dimod.SampleSet):
        labels = samples_like.variables
        samples = samples_like.record.sample
        num_occurrences = samples_like.record.num_occurrences
    else:
        samples, labels = dimod.as_samples(samples_like)
        num_occurrences = np.ones(samples.shape[0])

    if not all(v == idx for idx, v in enumerate(labels)):
        labels_to_idx = {v: idx for idx, v in enumerate(labels)}
        embedding = {v: {labels_to_idx[u] for u in chain} for v, chain in embedding.items()}

    if not embedding:
        return {}

    variables, chains = zip(*embedding.items())

    broken = broken_chains(samples, chains)

    return {v: float(np.average(broken[:, cidx], weights=num_occurrences))
            for cidx, v in enumerate(variables)}


def edgelist_to_adjacency(edgelist):
    """Converts an iterator of edges to an adjacency dict.

    Args:
        edgelist (iterable):
            An iterator over 2-tuples where each 2-tuple is an edge.

    Returns:
        dict: The adjacency dict. A dict of the form `{v: Nv, ...}` where `v` is
        a node in a graph and `Nv` is the neighbors of `v` as an set.

    """
    adjacency = dict()
    for u, v in edgelist:
        if u in adjacency:
            adjacency[u].add(v)
        else:
            adjacency[u] = {v}
        if v in adjacency:
            adjacency[v].add(u)
        else:
            adjacency[v] = {u}
    return adjacency

def adjacency_to_edges(adjacency):
    """Converts an adjacency dict, networkx graph, or bqm to an edge iterator.

    Args:
        adjacency (dict/:class:`networkx.Graph`/:class:`dimod.BQM`):
            Should be a dict of the form {s: Ns, ...} where s is a variable
            in the graph and Ns is the set of neighbours of s.

    Yields:
        tuple: A 2-tuple, corresponding to an edge in the provided graph

    """
    if hasattr(adjacency, 'edges'):
        yield from adjacency.edges()

    elif hasattr(adjacency, 'quadratic'):
        yield from adjacency.quadratic

    elif hasattr(adjacency, 'items'):
        seen = set()
        for v, Nv in adjacency.items():
            seen.add(v)
            for u in Nv:
                if u not in seen:
                    yield (u, v)
    else:
        raise TypeError("unrecognized type for adjacency -- provide a dict, "
                        "Mapping, networkx.Graph or dimod.BQM")

class intlabel_disjointsets:
    """A disjoint sets implementation with size and path-halving, for graphs 
    labeled [0, ..., n-1]

    Args:
        n (int):
            The number of items in the disjoint sets

    """
    def __init__(self, n):
        self._parent = list(range(n))
        self._size = [1] * n

    def find(self, q):
        """Find the current root for q.

        Args:
            q (int):
                A number in range(n)

        Returns:
            int: the root of the set containing q

        """
        parent = self._parent
        p = parent[q]
        while q != p:
            r = parent[q] = parent[p]
            q, p = p, r
        return p

    def union(self, p, q):
        """Merges the sets containing p and q.

        Args:
            p (int):
                A number in range(n)
            q (int):
                A number in range(n)

        """
        p = self.find(p)
        q = self.find(q)
        a = self._size[p]
        b = self._size[q]
        if p == q:
            return
        if a > b:
            p, q = q, p
        self._parent[p] = q
        self._size[q] = a + b

    def size(self, q):
        """Returns the size of the set containing q.

        Args:
            p (int):
                A number in range(n)

        Returns:
            int: the size of the set containing q
        """
        return self._size[self.find(q)]





# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

import networkx as nx

import dimod
import dwave.embedding
from dwave.embedding.chain_strength import uniform_torque_compensation, scaled

class TestUniformTorqueCompensation(unittest.TestCase):
    def setUp(self):
        h = {0: 0, 1: 0, 2: 0}
        J = {(0, 1): 1, (1, 2): 1, (0, 2): 1}
        self.bqm = dimod.BinaryQuadraticModel.from_ising(h, J)

    def test_empty(self):
        empty_bqm = dimod.BinaryQuadraticModel({}, {}, 0, 'SPIN')
        chain_strength = uniform_torque_compensation(empty_bqm, prefactor=2)
        self.assertEqual(chain_strength, 1)

    def test_default_prefactor(self):
        chain_strength = uniform_torque_compensation(self.bqm)
        self.assertAlmostEqual(chain_strength, 1.9997, places=4)

    def test_typical(self):
        chain_strength = uniform_torque_compensation(self.bqm, prefactor=2)
        self.assertAlmostEqual(chain_strength, 2.8284, places=4)

    def test_as_callable(self):
        embedding = {0: {0}, 1: {1}, 2: {2, 3}}
        embedded_bqm = dwave.embedding.embed_bqm(self.bqm, embedding, nx.cycle_graph(4), 
                                                 chain_strength=uniform_torque_compensation)

        expected_bqm = dimod.BinaryQuadraticModel({0: 0, 1: 0, 2: 0, 3: 0},
                                                  {(0, 1): 1, (1, 2): 1, (2, 3): -1.9997, (0, 3): 1},
                                                  1.9997,  # offset the energy from satisfying chains
                                                  dimod.SPIN)

        dimod.testing.assert_bqm_almost_equal(embedded_bqm, expected_bqm, places=4) 

class TestScaled(unittest.TestCase):
    def setUp(self):
        h = {0: 0, 1: -1, 2: 2}
        J = {(0, 1): 1, (1, 2): -2, (0, 2): -1}
        self.bqm = dimod.BinaryQuadraticModel.from_ising(h, J)

    def test_empty(self):
        empty_bqm = dimod.BinaryQuadraticModel({}, {}, 0, 'SPIN')
        chain_strength = scaled(empty_bqm, prefactor=2)
        self.assertEqual(chain_strength, 1)

    def test_default_prefactor(self):
        chain_strength = scaled(self.bqm)
        self.assertEqual(chain_strength, 2)

    def test_typical(self):
        chain_strength = scaled(self.bqm, prefactor=1.5)
        self.assertEqual(chain_strength, 3)

    def test_as_callable(self):
        embedding = {0: {0}, 1: {1}, 2: {2, 3}}
        embedded_bqm = dwave.embedding.embed_bqm(self.bqm, embedding, nx.cycle_graph(4), 
                                                 chain_strength=scaled)

        expected_bqm = dimod.BinaryQuadraticModel({0: 0, 1: -1, 2: 1, 3: 1},
                                                  {(0, 1): 1, (1, 2): -2, (2, 3): -2, (0, 3): -1},
                                                  2,  # offset the energy from satisfying chains
                                                  dimod.SPIN)
        self.assertEqual(embedded_bqm, expected_bqm)



# coding: utf-8
# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import warnings

import dimod

from dwave.system.cache.database_manager import get_flux_biases_from_cache, cache_connect, insert_flux_bias
from dwave.system.exceptions import MissingFluxBias
from dwave.system.samplers.dwave_sampler import DWaveSampler


def get_flux_biases(sampler, embedding, chain_strength, num_reads=1000, max_age=3600):
    """Get the flux bias offsets for sampler and embedding.

    Args:
        sampler (:obj:`.DWaveSampler`):
            A D-Wave sampler.

        embedding (dict[hashable, iterable]):
            Mapping from a source graph to the specified sampler’s graph (the target graph). The
            keys of embedding should be nodes in the source graph, the values should be an iterable
            of nodes in the target graph.

        chain_strength (number):
            Desired chain coupling strength. This is the magnitude of couplings between qubits
            in a chain.

        num_reads (int, optional, default=1000):
            The number of reads per system call if new flux biases need to be calculated.

        max_age (int, optional, default=3600):
            The maximum age (in seconds) allowed for previously calculated flux bias offsets.

    Returns:
        dict: A dict where the keys are the nodes in the chains and the values are the flux biases.

    """

    if not isinstance(sampler, dimod.Sampler):
        raise TypeError("input sampler should be DWaveSampler")

    # try to read the chip_id, otherwise get the name
    system_name = sampler.properties.get('chip_id', str(sampler.__class__))

    try:
        with cache_connect() as cur:
            fbo = get_flux_biases_from_cache(cur, embedding.values(), system_name,
                                             chain_strength=chain_strength,
                                             max_age=max_age)
        return fbo
    except MissingFluxBias:
        pass

    # if dwave-drivers is not available, then we can't calculate the biases
    try:
        import dwave.drivers as drivers
    except ImportError:
        msg = ("dwave-drivers not found, cannot calculate flux biases. dwave-drivers can be "
               "installed with "
               "'pip install dwave-drivers --extra-index-url https://pypi.dwavesys.com/simple'. "
               "See documentation for dwave-drivers license.")
        raise RuntimeError(msg)

    fbo = drivers.oneshot_flux_bias(sampler, embedding.values(), num_reads=num_reads,
                                    chain_strength=chain_strength,
                                    label='VirtualGraph flux bias measurements')

    # store them in the cache
    with cache_connect() as cur:
        for chain in embedding.values():
            v = next(iter(chain))
            flux_bias = fbo.get(v, 0.0)
            insert_flux_bias(cur, chain, system_name, flux_bias, chain_strength)

    return fbo



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
A :std:doc:`dimod sampler <oceandocs:docs_dimod/reference/samplers>` for the D-Wave system.

See :std:doc:`Ocean Glossary <oceandocs:glossary>`
for explanations of technical terms in descriptions of Ocean tools.
"""

import collections.abc as abc

import dimod

from dimod.exceptions import BinaryQuadraticModelStructureError
from dwave.cloud.client import Client
from dwave.cloud.exceptions import (
    SolverError, SolverAuthenticationError, InvalidAPIResponseError,
    RequestTimeout, PollingTimeout, ProblemUploadError, ProblemStructureError,
)

from dwave.system.exceptions import FailoverCondition, RetryCondition
from dwave.system.warnings import WarningHandler, WarningAction

import dwave_networkx as dnx
import networkx as nx

__all__ = ['DWaveSampler', 'qpu_graph']


def qpu_graph(topology_type, topology_shape, nodelist, edgelist):
    """Converts node and edge lists to a dwave-networkx compatible graph.

    Creates a D-Wave Chimera, Pegasus or Zephyr graph compatible with
    dwave-networkx libraries. 

    Args:
        topology_type (string):
            The type of lattice. Valid strings are `chimera`, `pegasus`
            and `zephyr`.
        topology_shape(iterable of ints):
            Specifies dimensions of the lattice.
        nodelist (list of ints):
            List of nodes in the graph. Node labeling is integer,
            and compatible with the topology_type linear labeling scheme.
        edgelist (list of Tuples):
            List of edges in the graph, each edge consisting of a pair
            of nodes.
    """
    
    if topology_type == 'chimera':
        if not (1 <= len(topology_shape) <=3):
            raise ValueError('topology_shape is incompatible with a chimera lattice.')
        G = dnx.chimera_graph(*topology_shape,
                              node_list=nodelist,
                              edge_list=edgelist)
    elif topology_type == 'pegasus':
        if len(topology_shape) != 1:
            raise ValueError('topology_shape is incompatible with a pegasus lattice.')
        G = dnx.pegasus_graph(topology_shape[0],
                                  node_list=nodelist,
                                  edge_list=edgelist)
    elif topology_type == 'zephyr':
        if len(topology_shape) not in (1, 2):
            raise ValueError('topology_shape is incompatible with a zephyr lattice.')
        G = dnx.zephyr_graph(*topology_shape,
                                 node_list=nodelist,
                                 edge_list=edgelist)
    else:
        # Alternative could be to create a standard network graph and
        # issue a warning. Requires new dependency on networkx.
        raise ValueError('topology_type does not match a known QPU architecure')
    return G


class DWaveSampler(dimod.Sampler, dimod.Structured):
    """A class for using the D-Wave system as a sampler for binary quadratic models.

    You can configure your :term:`solver` selection and usage by setting parameters,
    hierarchically, in a configuration file, as environment variables, or
    explicitly as input arguments. For more information, see
    `D-Wave Cloud Client <https://docs.ocean.dwavesys.com/en/stable/docs_cloud/sdk_index.html>`_
    :meth:`~dwave.cloud.client.Client.get_solvers`. By default, online
    D-Wave systems are returned ordered by highest number of qubits.

    Inherits from :class:`dimod.Sampler` and :class:`dimod.Structured`.

    Args:
        failover (bool, optional, default=False):
            Signal a failover condition if a sampling error occurs. When ``True``,
            raises :exc:`~dwave.system.exceptions.FailoverCondition` or
            :exc:`~dwave.system.exceptions.RetryCondition` on sampleset resolve
            to signal failover.

            Actual failover, i.e. selection of a new solver, has to be handled
            by the user. A convenience method :meth:`.trigger_failover` is available
            for this. Note that hardware graphs vary between QPUs, so triggering
            failover results in regenerated :attr:`.nodelist`, :attr:`.edgelist`,
            :attr:`.properties` and :attr:`.parameters`.

            .. versionchanged:: 1.16.0

               In the past, the :meth:`.sample` method was blocking and
               ``failover=True`` caused a solver failover and sampling retry.
               However, this failover implementation broke when :meth:`sample`
               became non-blocking (asynchronous), Setting ``failover=True`` had
               no effect.

        retry_interval (number, optional, default=-1):
            Ignored, but kept for backward compatibility.

            .. versionchanged:: 1.16.0

               Ignored since 1.16.0. See note for ``failover`` parameter above.

        **config:
            Keyword arguments passed to :meth:`dwave.cloud.client.Client.from_config`.

    Note:
        Prior to version 1.0.0, :class:`.DWaveSampler` used the ``base`` client,
        allowing non-QPU solvers to be selected.
        To reproduce the old behavior, instantiate :class:`.DWaveSampler` with
        ``client='base'``.

    Examples:
        This example submits a two-variable Ising problem mapped directly to two
        adjacent qubits on a D-Wave system. ``qubit_a`` is the first qubit in
        the QPU's indexed list of qubits and ``qubit_b`` is one of the qubits
        coupled to it. Other required parameters for communication with the system, such
        as its URL and an authentication token, are implicitly set in a configuration file
        or as environment variables, as described in
        `Configuring Access to D-Wave Solvers <https://docs.ocean.dwavesys.com/en/stable/overview/sapi.html>`_.
        Given sufficient reads (here 100), the quantum
        computer should return the best solution, :math:`{1, -1}` on ``qubit_a`` and
        ``qubit_b``, respectively, as its first sample (samples are ordered from
        lowest energy).

        >>> from dwave.system import DWaveSampler
        ...
        >>> sampler = DWaveSampler()
        ...
        >>> qubit_a = sampler.nodelist[0]
        >>> qubit_b = next(iter(sampler.adjacency[qubit_a]))
        >>> sampleset = sampler.sample_ising({qubit_a: -1, qubit_b: 1},
        ...                                  {},
        ...                                  num_reads=100)
        >>> sampleset.first.sample[qubit_a] == 1 and sampleset.first.sample[qubit_b] == -1
        True

    See `Ocean Glossary <https://docs.ocean.dwavesys.com/en/stable/concepts/index.html>`_
    for explanations of technical terms in descriptions of Ocean tools.

    """
    def __init__(self, failover=False, retry_interval=-1, **config):
        # strongly prefer QPU solvers; requires kwarg-level override
        config.setdefault('client', 'qpu')

        # weakly prefer QPU solver with the highest qubit count,
        # easily overridden on any config level above defaults (file/env/kwarg)
        defaults = config.setdefault('defaults', {})
        if not isinstance(defaults, abc.Mapping):
            raise TypeError("mapping expected for 'defaults'")
        defaults.update(solver=dict(order_by='-num_active_qubits'))

        self.client = Client.from_config(**config)
        self.solver = self.client.get_solver()

        self.failover = failover
        self.retry_interval = retry_interval

    warnings_default = WarningAction.IGNORE
    """Defines the default behavior for :meth:`.sample_ising`'s  and
    :meth:`sample_qubo`'s `warnings` kwarg.
    """

    @property
    def properties(self):
        """dict: D-Wave solver properties as returned by a SAPI query.

        Solver properties are dependent on the selected D-Wave solver and subject to change;
        for example, new released features may add properties.
        `D-Wave System Documentation <https://docs.dwavesys.com/docs/latest/doc_solver_ref.html>`_
        describes the parameters and properties supported on the D-Wave system.

        Examples:

            >>> from dwave.system import DWaveSampler
            >>> sampler = DWaveSampler()
            >>> sampler.properties    # doctest: +SKIP
            {'anneal_offset_ranges': [[-0.2197463755538704, 0.03821687759418928],
              [-0.2242514597680286, 0.01718456460967399],
              [-0.20860153999435985, 0.05511969218508182],
            # Snipped above response for brevity

        See `Ocean Glossary <https://docs.ocean.dwavesys.com/en/stable/concepts/index.html>`_
        for explanations of technical terms in descriptions of Ocean tools.

        """
        try:
            return self._properties
        except AttributeError:
            self._properties = properties = self.solver.properties.copy()
            return properties

    @property
    def parameters(self):
        """dict[str, list]: D-Wave solver parameters in the form of a dict, where keys are
        keyword parameters accepted by a SAPI query and values are lists of properties in
        :attr:`.properties` for each key.

        Solver parameters are dependent on the selected D-Wave solver and subject to change;
        for example, new released features may add parameters.
        `D-Wave System Documentation <https://docs.dwavesys.com/docs/latest/doc_solver_ref.html>`_
        describes the parameters and properties supported on the D-Wave system.

        Examples:

            >>> from dwave.system import DWaveSampler
            >>> sampler = DWaveSampler()
            >>> sampler.parameters    # doctest: +SKIP
            {'anneal_offsets': ['parameters'],
             'anneal_schedule': ['parameters'],
             'annealing_time': ['parameters'],
             'answer_mode': ['parameters'],
             'auto_scale': ['parameters'],
             # Snipped above response for brevity

        See `Ocean Glossary <https://docs.ocean.dwavesys.com/en/stable/concepts/index.html>`_
        for explanations of technical terms in descriptions of Ocean tools.

        """
        try:
            return self._parameters
        except AttributeError:
            parameters = {param: ['parameters']
                          for param in self.properties['parameters']}
            parameters.update(warnings=[])
            parameters.update(label=[])
            self._parameters = parameters
            return parameters

    @property
    def edgelist(self):
        """list: List of active couplers for the D-Wave solver.

        Examples:
            Coupler list for one D-Wave 2000Q system (output snipped for brevity).

            >>> from dwave.system import DWaveSampler
            >>> sampler = DWaveSampler()
            >>> sampler.edgelist
            [(0, 4), (0, 5), (0, 6), (0, 7), ...

        See `Ocean Glossary <https://docs.ocean.dwavesys.com/en/stable/concepts/index.html>`_
        for explanations of technical terms in descriptions of Ocean tools.

        """
        # Assumption: cloud client nodes are always integer-labelled
        try:
            edgelist = self._edgelist
        except AttributeError:
            self._edgelist = edgelist = sorted(set((u, v) if u < v else (v, u)
                                                   for u, v in self.solver.edges))
        return edgelist

    @property
    def nodelist(self):
        """list: List of active qubits for the D-Wave solver.

        Examples:
            Node list for one D-Wave 2000Q system (output snipped for brevity).

            >>> from dwave.system import DWaveSampler
            >>> sampler = DWaveSampler()
            >>> sampler.nodelist
            [0, 1, 2, ...

        See `Ocean Glossary <https://docs.ocean.dwavesys.com/en/stable/concepts/index.html>`_
        for explanations of technical terms in descriptions of Ocean tools.

        """
        # Assumption: cloud client nodes are always integer-labelled
        try:
            nodelist = self._nodelist
        except AttributeError:
            self._nodelist = nodelist = sorted(self.solver.nodes)
        return nodelist

    def trigger_failover(self):
        """Trigger a failover and connect to a new solver."""

        # the requested features are saved on the client object, so
        # we just need to request a new solver
        self.solver = self.client.get_solver(refresh=True)

        # delete the lazily-constructed attributes
        try:
            del self._edgelist
        except AttributeError:
            pass

        try:
            del self._nodelist
        except AttributeError:
            pass

        try:
            del self._parameters
        except AttributeError:
            pass

        try:
            del self._properties
        except AttributeError:
            pass

    def sample(self, bqm, warnings=None, **kwargs):
        """Sample from the specified binary quadratic model.

        Args:
            bqm (:class:`~dimod.BinaryQuadraticModel`):
                The binary quadratic model. Must match :attr:`.nodelist` and
                :attr:`.edgelist`.

            warnings (:class:`~dwave.system.warnings.WarningAction`, optional):
                Defines what warning action to take, if any. See
                :ref:`warnings_system`. The default behaviour is to
                ignore warnings.

            **kwargs:
                Optional keyword arguments for the sampling method, specified per solver in
                :attr:`.parameters`. D-Wave System Documentation's
                `solver guide <https://docs.dwavesys.com/docs/latest/doc_solver_ref.html>`_
                describes the parameters and properties supported on the D-Wave system.

        Returns:
            :class:`~dimod.SampleSet`: Sample set constructed from a (non-blocking)
            :class:`~concurrent.futures.Future`-like object.
            In it this sampler also provides timing information in the `info`
            field as described in the D-Wave System Documentation's
            :ref:`sysdocs_gettingstarted:qpu_sapi_qpu_timing`.

        Examples:
            This example submits a two-variable Ising problem mapped directly to two
            adjacent qubits on a D-Wave system. ``qubit_a`` is the first qubit in
            the QPU's indexed list of qubits and ``qubit_b`` is one of the qubits
            coupled to it. Given sufficient reads (here 100), the quantum
            computer should return the best solution, :math:`{1, -1}` on ``qubit_a`` and
            ``qubit_b``, respectively, as its first sample (samples are ordered from
            lowest energy).

            >>> from dwave.system import DWaveSampler
            ...
            >>> sampler = DWaveSampler()
            ...
            >>> qubit_a = sampler.nodelist[0]
            >>> qubit_b = next(iter(sampler.adjacency[qubit_a]))
            >>> sampleset = sampler.sample_ising({qubit_a: -1, qubit_b: 1},
            ...                                  {},
            ...                                  num_reads=100)
            >>> sampleset.first.sample[qubit_a] == 1 and sampleset.first.sample[qubit_b] == -1
            True

        See `Ocean Glossary <https://docs.ocean.dwavesys.com/en/stable/concepts/index.html>`_
        for explanations of technical terms in descriptions of Ocean tools.

        """

        solver = self.solver

        try:
            future = solver.sample_bqm(bqm, **kwargs)
        except ProblemStructureError as exc:
            msg = ("Problem graph incompatible with solver. Please use 'EmbeddingComposite' "
                   "to map the problem graph to the solver.")
            raise BinaryQuadraticModelStructureError(msg) from exc

        if warnings is None:
            warnings = self.warnings_default
        warninghandler = WarningHandler(warnings)
        warninghandler.energy_scale(bqm)

        # need a hook so that we can lazily check the sampleset for warnings
        # and handle failover consistently
        def _hook(computation):
            def resolve(computation):
                sampleset = computation.sampleset
                sampleset.resolve()

                if warninghandler is not None:
                    warninghandler.too_few_samples(sampleset)
                    if warninghandler.action is WarningAction.SAVE:
                        sampleset.info['warnings'] = warninghandler.saved

                return sampleset

            try:
                return resolve(computation)

            except (ProblemUploadError, RequestTimeout, PollingTimeout) as exc:
                if not self.failover:
                    raise exc

                # failover with retry on:
                # - request or polling timeout
                # - upload errors
                raise RetryCondition("resubmit problem") from exc

            except (SolverError, InvalidAPIResponseError) as exc:
                if not self.failover:
                    raise exc
                if isinstance(exc, SolverAuthenticationError):
                    raise exc

                # failover on:
                # - solver offline, solver disabled or not found
                # - internal SAPI errors (like malformed response)
                # - generic solver errors
                # but NOT on auth errors
                raise FailoverCondition("switch solver and resubmit problem") from exc

        return dimod.SampleSet.from_future(future, _hook)

    def sample_ising(self, h, *args, **kwargs):
        # to be consistent with the cloud-client, we ignore the 0 biases
        # on missing nodes for lists
        if isinstance(h, list):
            if len(h) > self.solver.num_qubits:
                msg = ("Problem graph incompatible with solver. Please use 'EmbeddingComposite' "
                       "to map the problem graph to the solver.")
                raise BinaryQuadraticModelStructureError(msg)
            nodes = self.solver.nodes
            h = dict((v, b) for v, b in enumerate(h) if b and v in nodes)

        return super().sample_ising(h, *args, **kwargs)

    def validate_anneal_schedule(self, anneal_schedule):
        """Raise an exception if the specified schedule is invalid for the sampler.

        Args:
            anneal_schedule (list):
                An anneal schedule variation is defined by a series of pairs of floating-point
                numbers identifying points in the schedule at which to change slope. The first
                element in the pair is time t in microseconds; the second, normalized persistent
                current s in the range [0,1]. The resulting schedule is the piecewise-linear curve
                that connects the provided points.

        Raises:
            ValueError: If the schedule violates any of the conditions listed below.

            RuntimeError: If the sampler does not accept the `anneal_schedule` parameter or
                if it does not have `annealing_time_range` or `max_anneal_schedule_points`
                properties.

        As described in
        `D-Wave System Documentation <https://docs.dwavesys.com/docs/latest/doc_solver_ref.html>`_,
        an anneal schedule must satisfy the following conditions:

            * Time t must increase for all points in the schedule.
            * For forward annealing, the first point must be (0,0) and the anneal fraction s must
              increase monotonically.
            * For reverse annealing, the anneal fraction s must start and end at s=1.
            * In the final point, anneal fraction s must equal 1 and time t must not exceed the
              maximum  value in the `annealing_time_range` property.
            * The number of points must be >=2.
            * The upper bound is system-dependent; check the `max_anneal_schedule_points` property.
              For reverse annealing, the maximum number of points allowed is one more than the
              number given by this property.

        Examples:
            This example sets a quench schedule on a D-Wave system.

            >>> from dwave.system import DWaveSampler
            >>> sampler = DWaveSampler()
            >>> quench_schedule=[[0.0, 0.0], [12.0, 0.6], [12.8, 1.0]]
            >>> DWaveSampler().validate_anneal_schedule(quench_schedule)    # doctest: +SKIP
            >>>

        """
        if 'anneal_schedule' not in self.parameters:
            raise RuntimeError("anneal_schedule is not an accepted parameter for this sampler")

        properties = self.properties

        try:
            min_anneal_time, max_anneal_time = properties['annealing_time_range']
            max_anneal_schedule_points = properties['max_anneal_schedule_points']
        except KeyError:
            raise RuntimeError("annealing_time_range and max_anneal_schedule_points are not properties of this solver")

        # The number of points must be >= 2.
        # The upper bound is system-dependent; check the max_anneal_schedule_points property
        if not isinstance(anneal_schedule, list):
            raise TypeError("anneal_schedule should be a list")
        elif len(anneal_schedule) < 2 or len(anneal_schedule) > max_anneal_schedule_points:
            msg = ("anneal_schedule must contain between 2 and {} points (contains {})"
                   ).format(max_anneal_schedule_points, len(anneal_schedule))
            raise ValueError(msg)

        try:
            t_list, s_list = zip(*anneal_schedule)
        except ValueError:
            raise ValueError("anneal_schedule should be a list of 2-tuples")

        # Time t must increase for all points in the schedule.
        if not all(tail_t < lead_t for tail_t, lead_t in zip(t_list, t_list[1:])):
            raise ValueError("Time t must increase for all points in the schedule")

        # max t cannot exceed max_anneal_time
        if t_list[-1] > max_anneal_time:
            raise ValueError("schedule cannot be longer than the maximum anneal time of {}".format(max_anneal_time))

        start_s, end_s = s_list[0], s_list[-1]
        if end_s != 1:
            raise ValueError("In the final point, anneal fraction s must equal 1.")
        if start_s == 1:
            # reverse annealing
            pass
        elif start_s == 0:
            # forward annealing, s must monotonically increase.
            if not all(tail_s <= lead_s for tail_s, lead_s in zip(s_list, s_list[1:])):
                raise ValueError("For forward anneals, anneal fraction s must monotonically increase")
        else:
            msg = ("In the first point, anneal fraction s must equal 0 for forward annealing or "
                   "1 for reverse annealing")
            raise ValueError(msg)

        # finally check the slope abs(slope) < 1/min_anneal_time
        max_slope = 1.0 / min_anneal_time
        for (t0, s0), (t1, s1) in zip(anneal_schedule, anneal_schedule[1:]):
            if round(abs((s0 - s1) / (t0 - t1)),10) > max_slope:
                raise ValueError("the maximum slope cannot exceed {}".format(max_slope))

    
        
    
    def to_networkx_graph(self):
        """Converts DWaveSampler's structure to a Chimera, Pegasus or Zephyr NetworkX graph.

        Returns:
            :class:`networkx.Graph`:
                Either a Chimera lattice of shape [m, n, t], a Pegasus 
                lattice of shape [m] or a Zephyr lattice of size [m,t].

        Examples:
            This example converts a selected D-Wave system solver to a graph
            and verifies it has over 2000 nodes.

            >>> from dwave.system import DWaveSampler
            ...
            >>> sampler = DWaveSampler()
            >>> g = sampler.to_networkx_graph()      # doctest: +SKIP
            >>> len(g.nodes) > 2000                  # doctest: +SKIP
            True
        """
        return qpu_graph(self.properties['topology']['type'],
                         self.properties['topology']['shape'],
                         self.nodelist, self.edgelist)
        



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest
import time
import homebase

import networkx as nx

from dwave.system.exceptions import MissingFluxBias
import dwave.system.cache as cache

tmp_database_name = 'tmp_test_database_manager_{}.db'.format(time.time())
# test_database_path = cache.cache_file(filename=tmp_database_name)
# conn = cache.cache_connect(test_database_path)


class TestCacheManager(unittest.TestCase):
    def test_same_database(self):
        """multiple calls to get_database_path"""
        db1 = cache.cache_file()
        db2 = cache.cache_file()

        self.assertEqual(db1, db2)


class TestDatabaseManager(unittest.TestCase):
    def setUp(self):
        # new connection for just this test
        self.clean_connection = cache.cache_connect(':memory:')

        # test_database_path = cache.cache_file(filename=tmp_database_name)
        # test_connection = cache.cache_connect(test_database_path)

    def tearDown(self):
        # close the memory connection
        self.clean_connection.close()

    def test_insert_retrieve_chain(self):
        """insert and retrieve some chains"""
        conn = self.clean_connection

        chain = [0, 1, 3]

        # insert the chain twice, should only result in one net insert
        with conn as cur:
            cache.insert_chain(cur, chain)
            cache.insert_chain(cur, chain)

        with conn as cur:
            all_chains = list(cache.iter_chain(cur))
        self.assertEqual(len(all_chains), 1)
        returned_chain, = all_chains
        self.assertEqual(len(returned_chain), len(chain))
        self.assertEqual(returned_chain, chain)

        # insert a new chain
        with conn as cur:
            cache.insert_chain(cur, [0, 2, 3])
        with conn as cur:
            all_chains = list(cache.iter_chain(cur))
        self.assertEqual(len(all_chains), 2)  # should now be two chains

    def test_insert_flux_bias(self):
        conn = self.clean_connection

        with conn as cur:
            cache.insert_flux_bias(cur, [0, 1, 2], 'test_system', .1, 1)

        with conn as cur:
            # try dumping
            flux_biases = list(cache.iter_flux_bias(cur))
        self.assertEqual(len(flux_biases), 1)
        self.assertEqual(flux_biases[0], ([0, 1, 2], 'test_system', .1, 1))

        # retrieve by name
        with conn as cur:
            biases = cache.get_flux_biases_from_cache(cur, [[0, 1, 2]], 'test_system', 1)

        for v, fbo in biases.items():
            self.assertIn(v, [0, 1, 2])
            self.assertEqual(fbo, .1)

        # now get something wrong out
        with self.assertRaises(MissingFluxBias):
            with conn as cur:
                biases = cache.get_flux_biases_from_cache(cur, [[0, 1, 2]], 'another_system', 1)

    def test_graph_insert_retrieve(self):
        conn = self.clean_connection

        graph = nx.barbell_graph(8, 8)
        nodelist = sorted(graph)
        edgelist = sorted(sorted(edge) for edge in graph.edges)

        with conn as cur:
            cache.insert_graph(cur, nodelist, edgelist)

            # should only be one graph
            graphs = list(cache.iter_graph(cur))
            self.assertEqual(len(graphs), 1)
            (nodelist_, edgelist_), = graphs
            self.assertEqual(nodelist, nodelist_)
            self.assertEqual(edgelist, edgelist_)

        # trying to reinsert should still result in only one graph
        with conn as cur:
            cache.insert_graph(cur, nodelist, edgelist)
            graphs = list(cache.iter_graph(cur))
            self.assertEqual(len(graphs), 1)

        # inserting with an empty dict as encoded_data should populate it
        encoded_data = {}
        with conn as cur:
            cache.insert_graph(cur, nodelist, edgelist, encoded_data)
        self.assertIn('num_nodes', encoded_data)
        self.assertIn('num_edges', encoded_data)
        self.assertIn('edges', encoded_data)

        # now adding another graph should result in two items
        graph = nx.complete_graph(4)
        nodelist = sorted(graph)
        edgelist = sorted(sorted(edge) for edge in graph.edges)
        with conn as cur:
            cache.insert_graph(cur, nodelist, edgelist)
            graphs = list(cache.iter_graph(cur))
            self.assertEqual(len(graphs), 2)

    def test_insert_embedding(self):
        test_database_path = cache.cache_file(filename=tmp_database_name)
        conn = cache.cache_connect(test_database_path)

        source_nodes = [0, 1, 2]
        source_edges = [[0, 1], [0, 2], [1, 2]]
        target_nodes = [0, 1, 2, 3]
        target_edges = [[0, 1], [0, 3], [1, 2], [2, 3]]

        embedding = {0: [0], 1: [1], 2: [2, 3]}

        with conn as cur:
            cache.insert_embedding(cur, source_nodes, source_edges, target_nodes, target_edges,
                                     embedding, 'tag1')

            embedding_ = cache.select_embedding_from_tag(cur, 'tag1', target_nodes, target_edges)

            self.assertEqual(embedding, embedding_)

        # now reinsert but with a different embedding
        embedding = {0: [0, 1], 1: [2], 2: [3]}
        with conn as cur:
            cache.insert_embedding(cur, source_nodes, source_edges, target_nodes, target_edges,
                                     embedding, 'tag1')

            # get it back
            embedding_ = cache.select_embedding_from_tag(cur, 'tag1', target_nodes, target_edges)

            self.assertEqual(embedding, embedding_)

            # get it back from source graph
            embedding_ = cache.select_embedding_from_source(cur, source_nodes, source_edges,
                                                              target_nodes, target_edges)
            self.assertEqual(embedding, embedding_)



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import os
import json
import unittest

from parameterized import parameterized_class

from dwave.cloud.exceptions import ConfigFileError, SolverNotFoundError
from dwave.cloud.testing import isolated_environ
from dwave.system import LeapHybridSampler, LeapHybridDQMSampler, LeapHybridCQMSampler


@parameterized_class(
    ("problem_type", "sampler_cls"), [
        ("bqm", LeapHybridSampler),
        ("dqm", LeapHybridDQMSampler),
        ("cqm", LeapHybridCQMSampler),
    ])
@unittest.skipIf(os.getenv('SKIP_INT_TESTS'), "Skipping integration test.")
class TestLegacySolverSelection(unittest.TestCase):

    def test_baseline(self):
        # kwarg-level override should always work
        try:
            sampler = self.sampler_cls(solver=self.sampler_cls.default_solver)
        except (ValueError, ConfigFileError):
            raise unittest.SkipTest("config error")

        self.assertIn(self.problem_type, sampler.solver.supported_problem_types)

        sampler.client.close()

    def test_new_positive(self):
        # given valid solver spec in env, init succeeds
        solver = json.dumps(self.sampler_cls.default_solver)
        with isolated_environ(add={'DWAVE_API_SOLVER': solver},
                              remove=('DWAVE_FEATURE_FLAGS',)):
            try:
                sampler = self.sampler_cls()
            except (ValueError, ConfigFileError):
                raise unittest.SkipTest("config error")

            self.assertIn(self.problem_type, sampler.solver.supported_problem_types)

        sampler.client.close()

    def test_new_negative(self):
        # given invalid solver spec in env, init fails
        solver = json.dumps(dict(qpu=True))
        with isolated_environ(add={'DWAVE_API_SOLVER': solver},
                              remove=('DWAVE_FEATURE_FLAGS',)):
            with self.assertRaises(SolverNotFoundError):
                try:
                    sampler = self.sampler_cls()
                except (ValueError, ConfigFileError):
                    raise unittest.SkipTest("config error")

    def test_old(self):
        # given invalid solver spec in env, init succeeds
        solver = json.dumps(dict(qpu=True))
        flags = json.dumps(dict(hss_solver_config_override=True))
        with isolated_environ(add={'DWAVE_API_SOLVER': solver,
                                   'DWAVE_FEATURE_FLAGS': flags}):
            try:
                sampler = self.sampler_cls()
            except (ValueError, ConfigFileError):
                raise unittest.SkipTest("config error")

            self.assertIn(self.problem_type, sampler.solver.supported_problem_types)

        sampler.client.close()



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

import dimod
import dimod.testing as dtest

from dwave.system import CutOffComposite


class CutoffChecker(dimod.Sampler):
    def __init__(self, child_sampler, expected_bqm):

        self.child = child_sampler
        self.bqm = expected_bqm

    def sample(self, bqm, **parameters):
        assert self.bqm == bqm, '{} != {}'.format(self.bqm, bqm)
        return self.child.sample(bqm, **parameters)

    def parameters(self):
        return self.child.parameters()

    def properties(self):
        return self.child.properties()


class TestConstruction(unittest.TestCase):
    def test_instantiation_smoketest(self):
        sampler = CutOffComposite(dimod.ExactSolver(), 0)
        dtest.assert_sampler_api(sampler)


class TestCutoffIsing(unittest.TestCase):
    def test_no_cutoff(self):
        h = {'a': -4.0, 'b': -4.0}
        J = {('a', 'b'): 3.2, ('b', 'c'): 0.1}
        cutoff = -1

        bqm = dimod.BinaryQuadraticModel.from_ising(h, J)

        checker = CutoffChecker(dimod.ExactSolver(), bqm)
        samples = CutOffComposite(checker, cutoff).sample_ising(h, J)

        dimod.testing.assert_response_energies(samples, bqm)

    def test_triange_to_3path(self):
        h = {}
        J = {'ab': -1, 'bc': -1, 'ac': .5}
        cutoff = .75
        cut = dimod.BinaryQuadraticModel.from_ising({}, {'ab': -1, 'bc': -1})

        checker = CutoffChecker(dimod.ExactSolver(), cut)
        samples = CutOffComposite(checker, cutoff).sample_ising(h, J)

        dimod.testing.assert_response_energies(samples, dimod.BinaryQuadraticModel.from_ising(h, J))

    def test_triangle_to_2path(self):
        h = {'a': 1.2, 'b': 1, 'c': .5}
        J = {'ab': -1, 'bc': -.5, 'ac': -.5}
        cutoff = .75
        cut = dimod.BinaryQuadraticModel.from_ising({'a': 1.2, 'b': 1}, {'ab': -1})

        checker = CutoffChecker(dimod.ExactSolver(), cut)
        samples = CutOffComposite(checker, cutoff).sample_ising(h, J)

        dimod.testing.assert_response_energies(samples, dimod.BinaryQuadraticModel.from_ising(h, J))

        # check that we picked the right minimizing value for the isolated
        self.assertEqual(samples.first.sample['c'], -1)

    def test_triangle_to_empty(self):
        h = {'a': 1.2, 'b': 1, 'c': .5}
        J = {'ab': -.5, 'bc': -.5, 'ac': -.5}
        cutoff = .75
        cut = dimod.BinaryQuadraticModel.from_ising({}, {})

        # we cannot check in this case because all variables are isolated
        # this results in exactly one variable being sent to ExactSolver and
        # we don't know which one it will be, so we just check the correctness
        # of the output
        samples = CutOffComposite(dimod.ExactSolver(), cutoff).sample_ising(h, J)

        dimod.testing.assert_response_energies(samples, dimod.BinaryQuadraticModel.from_ising(h, J))

        # check that we picked the right minimizing value for the isolated
        self.assertEqual(samples.first.sample['a'], -1)
        self.assertEqual(samples.first.sample['b'], -1)
        self.assertEqual(samples.first.sample['c'], -1)

    def test_4_path_isolated_tail(self):
        h = {}
        J = {'ab': -1, 'bc': -.5, 'cd': -.5, 'de': -.5}
        cutoff = .75
        cut = dimod.BinaryQuadraticModel.from_ising({}, {'ab': -1})

        checker = CutoffChecker(dimod.ExactSolver(), cut)
        samples = CutOffComposite(checker, cutoff).sample_ising(h, J)

        dimod.testing.assert_response_energies(samples, dimod.BinaryQuadraticModel.from_ising(h, J))



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import os
import unittest

import dimod

from dwave.cloud.exceptions import ConfigFileError, SolverNotFoundError
from dwave.system import LeapHybridSampler

try:
    sampler = LeapHybridSampler()
except (ValueError, ConfigFileError, SolverNotFoundError):
    sampler = None


@unittest.skipIf(os.getenv('SKIP_INT_TESTS'), "Skipping integration test.")
@unittest.skipIf(sampler is None, "no hybrid sampler available")
# @dimod.testing.load_sampler_bqm_tests(sampler)  # these take a while
class TestLeapHybridSampler(unittest.TestCase):
    def test_smoke(self):
        sampleset = sampler.sample_ising({'a': -1}, {'ab': 1})
        sampleset.resolve()

    # NOTE: enable when problem labelling deployed to prod
    @unittest.skipIf(sampler is not None and 'cloud' in sampler.client.endpoint,
                     "labels not supported in production")
    def test_problem_labelling(self):
        bqm = dimod.BQM.from_ising({'a': -1}, {'ab': 1})
        label = 'problem label'

        # label set
        sampleset = sampler.sample(bqm, label=label)
        self.assertIn('problem_id', sampleset.info)
        self.assertIn('problem_label', sampleset.info)
        self.assertEqual(sampleset.info['problem_label'], label)

        # label unset
        sampleset = sampler.sample(bqm)
        self.assertIn('problem_id', sampleset.info)
        self.assertNotIn('problem_label', sampleset.info)



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import unittest

import dimod

from dwave.cloud.exceptions import ConfigFileError, SolverNotFoundError
from dwave.system import LeapHybridCQMSampler

try:
    sampler = LeapHybridCQMSampler()
except (ValueError, ConfigFileError, SolverNotFoundError):
    sampler = None


@unittest.skipIf(os.getenv('SKIP_INT_TESTS'), "Skipping integration test.")
@unittest.skipIf(sampler is None, "no hybrid sampler available")
class TestLeapHybridSampler(unittest.TestCase):

    def assertConsistent(self, cqm, sampleset):
        self.assertEqual(cqm.variables, sampleset.variables)

        # todo: check the vartypes etc

    def test_small(self):
        # submit 3-variable CQMs of binary, spin, integer and mixed
        objectives = dict()

        objectives['binary'] = dimod.BQM({'a': 1, 'b': 1, 'c': 1}, {}, 0, 'BINARY')
        objectives['spin'] = dimod.BQM({'a': 1, 'b': 1, 'c': 1}, {}, 0, 'SPIN')

        objectives['integer'] = integer = dimod.QM()
        integer.add_variables_from('INTEGER', 'abc')

        objectives['mixed'] = mixed = dimod.QM()
        mixed.add_variable('BINARY', 'a')
        mixed.add_variable('SPIN', 'b')
        mixed.add_variable('INTEGER', 'c')

        for vartype, model in objectives.items():
            with self.subTest(f"one constraint, vartype={vartype}"):
                cqm = dimod.ConstrainedQuadraticModel()
                cqm.set_objective(model)
                cqm.add_constraint(model, rhs=1, sense='==')
                sampleset = sampler.sample_cqm(cqm)
                self.assertConsistent(cqm, sampleset)

            with self.subTest(f"no constraints, vartype={vartype}"):
                cqm = dimod.ConstrainedQuadraticModel()
                cqm.set_objective(model)
                sampleset = sampler.sample_cqm(cqm)
                self.assertConsistent(cqm, sampleset)

    def test_large(self):
        # submit large CQMs of binary, spin, integer and mixed
        num_variables = 5000

        objectives = dict()

        objectives['binary'] = dimod.BQM(num_variables, 'BINARY')
        objectives['spin'] = dimod.BQM(num_variables, 'SPIN')

        objectives['integer'] = integer = dimod.QM()
        integer.add_variables_from('INTEGER', range(num_variables))

        objectives['mixed'] = mixed = dimod.QM()
        mixed.add_variables_from('INTEGER', range(num_variables // 3))
        mixed.add_variables_from(
            'SPIN', range(mixed.num_variables, mixed.num_variables + (num_variables // 3)))
        mixed.add_variables_from('BINARY', range(mixed.num_variables, num_variables))

        for vartype, model in objectives.items():
            with self.subTest(f"one constraint, vartype={vartype}"):
                cqm = dimod.ConstrainedQuadraticModel()
                cqm.set_objective(model)
                cqm.add_constraint(model, rhs=1, sense='==')
                sampleset = sampler.sample_cqm(cqm)
                self.assertConsistent(cqm, sampleset)

            with self.subTest(f"no constraints, vartype={vartype}"):
                cqm = dimod.ConstrainedQuadraticModel()
                cqm.set_objective(model)
                sampleset = sampler.sample_cqm(cqm)
                self.assertConsistent(cqm, sampleset)

    def test_linear_constraints(self):
        num_constraints = 100000

        x = dimod.Binary('x')

        cqm = dimod.ConstrainedQuadraticModel()
        for _ in range(num_constraints):
            cqm.add_constraint(x == 1)

        sampler.sample_cqm(cqm).resolve()  # smoke test

    def test_other_quadratic_model(self):
        with self.assertRaises(TypeError):
            sampler.sample_cqm(dimod.BQM('SPIN'))

    def test_quadratic_constraints(self):
        num_constraints = 100000

        x = dimod.Binary('x')
        y = dimod.Binary('y')

        cqm = dimod.ConstrainedQuadraticModel()
        for _ in range(num_constraints):
            cqm.add_constraint(x*y == 1)

        sampler.sample_cqm(cqm).resolve()  # smoke test



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

from unittest.mock import patch

import dimod

from dwave.system import LeapHybridDQMSampler


class TestLeapHybridDQMSampler(unittest.TestCase):
    def test_time_limit_exceptions(self):

        class MockSolver():
            properties = dict(category='hybrid',
                              minimum_time_limit=[[20000, 5.0],
                                                  [100000, 6.0],
                                                  [200000, 13.0],
                                                  [500000, 34.0],
                                                  [1000000, 71.0],
                                                  [2000000, 152.0],
                                                  [5000000, 250.0],
                                                  [20000000, 400.0],
                                                  [250000000, 1200.0]],
                              maximum_time_limit_hrs=24.0,
                              )
            supported_problem_types = ['dqm']

            def sample_dqm(self, *args, **kwargs):
                raise RuntimeError

        class MockClient():
            @classmethod
            def from_config(cls, *args, **kwargs):
                return cls()

            def get_solver(self, *args, **kwargs):
                return MockSolver()

        with patch('dwave.system.samplers.leap_hybrid_sampler.Client', MockClient):
            sampler = LeapHybridDQMSampler()

            dqm = dimod.DQM()

        with self.assertRaises(ValueError):
            sampler.sample_dqm(dqm, time_limit=1)

        with self.assertRaises(ValueError):
            sampler.sample_dqm(dqm, time_limit=10000000)

    def test_DQM_subclass_without_serialization_can_be_sampled(self):
        """Test that DQM subclasses that do not implement serialization can
        still be sampled by LeapHybridDQMSampler.

        Sampling requires serialization.  LeapHybridDQMSampler calls
        DQM.to_file(dqm, ...) if dqm.to_file(...) raises NotImplementedError.
        """

        class DQMWithoutSerialization(dimod.DQM):
            def to_file(self, *args, **kwargs):
                raise NotImplementedError

        class MockSampleset:
            def relabel_variables(self, *args):
                return self

        class MockFuture:
            sampleset = MockSampleset()

        class MockSolver:
            properties = dict(category='hybrid',
                              minimum_time_limit=[[10000, 1.0]]
                              )
            supported_problem_types = ['dqm']

            def sample_dqm(self, *args, **kwargs):
                return MockFuture()

        class MockClient:
            @classmethod
            def from_config(cls, *args, **kwargs):
                return cls()

            def get_solver(self, *args, **kwargs):
                return MockSolver()

        with patch('dwave.system.samplers.leap_hybrid_sampler.Client', MockClient):
            sampler = LeapHybridDQMSampler()

            dqm = DQMWithoutSerialization()

        sampler.sample_dqm(dqm)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================
"""The schema used by the sqlite database for storing the penalty models."""

schema = \
    """
    CREATE TABLE IF NOT EXISTS chain(
        nodes TEXT UNIQUE,  -- json list of nodes in the chain
        chain_length INTEGER NOT NULL,  -- number of nodes in the chain
        id INTEGER PRIMARY KEY);

    CREATE TABLE IF NOT EXISTS system(
        system_name TEXT UNIQUE,  -- the name of the system
        id INTEGER PRIMARY KEY);

    CREATE TABLE IF NOT EXISTS flux_bias(
        chain_id INTEGER NOT NULL,
        system_id INTEGER NOT NULL,
        insert_time DATE NOT NULL,  -- flux biases are only valid for relatively short time
        flux_bias BLOB NOT NULL,
        chain_strength BLOB NOT NULL,  -- magnitude of the chain strength
        FOREIGN KEY (chain_id) REFERENCES chain(id) ON DELETE CASCADE,
        FOREIGN KEY (system_id) REFERENCES system(id) ON DELETE CASCADE,
        PRIMARY KEY(chain_id, system_id, chain_strength));

    CREATE VIEW IF NOT EXISTS flux_bias_view AS
        SELECT
            chain.chain_length,
            chain.nodes,
            system.system_name,
            flux_bias.flux_bias,
            flux_bias.insert_time,
            flux_bias.chain_strength
        FROM flux_bias, chain, system WHERE
            flux_bias.system_id = system.id AND
            flux_bias.chain_id = chain.id;

    CREATE TABLE IF NOT EXISTS graph(
        num_nodes INTEGER NOT NULL,  -- for integer-labeled graphs, num_nodes encodes all of the nodes
        num_edges INTEGER NOT NULL,  -- redundant, allows for faster selects
        edges TEXT NOT NULL,  -- json list of lists, should be sorted (with each edge sorted)
        id INTEGER PRIMARY KEY,
        CONSTRAINT graph UNIQUE (
            num_nodes,
            edges));

    CREATE TABLE IF NOT EXISTS embedding(
        source_id INTEGER NOT NULL,
        target_id INTEGER NOT NULL,
        tag TEXT,  -- user specified
        id INTEGER PRIMARY KEY,
        protected INTEGER DEFAULT 0,
        CONSTRAINT embedding_constraint UNIQUE (target_id, tag),
        CONSTRAINT embedding_constraint UNIQUE (target_id, source_id),
        FOREIGN KEY (source_id) REFERENCES graph(id) ON DELETE CASCADE,
        FOREIGN KEY (target_id) REFERENCES graph(id) ON DELETE CASCADE);

    CREATE TABLE IF NOT EXISTS embedding_component(
        source_node INTEGER NOT NULL,
        chain_id INTEGER NOT NULL,
        embedding_id INTEGER NOT NULL,
        FOREIGN KEY (chain_id) REFERENCES chain(id) ON DELETE CASCADE,
        FOREIGN KEY (embedding_id) REFERENCES embedding(id) ON DELETE CASCADE,
        PRIMARY KEY(embedding_id, source_node));

    CREATE VIEW IF NOT EXISTS embedding_component_view AS
        SELECT
            source_graph.num_nodes 'source_num_nodes',
            source_graph.num_edges 'source_num_edges',
            source_graph.edges 'source_edges',

            target_graph.num_nodes 'target_num_nodes',
            target_graph.num_edges 'target_num_edges',
            target_graph.edges 'target_edges',

            embedding_component.source_node,

            chain.chain_length,
            chain.nodes 'chain',

            embedding.tag 'embedding_tag'
        FROM
            graph 'source_graph',
            graph 'target_graph',
            embedding,
            embedding_component,
            chain

        WHERE
            embedding.source_id = source_graph.id AND
            embedding.target_id = target_graph.id AND
            embedding_component.chain_id = chain.id AND
            embedding.id = embedding_component.embedding_id;
    """



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import itertools
import unittest

import minorminer
import dimod.testing as dtest

from dwave.cloud.exceptions import ConfigFileError, SolverNotFoundError

from dwave.system.composites import VirtualGraphComposite
from dwave.system.samplers import DWaveSampler


@unittest.skipIf(os.getenv('SKIP_INT_TESTS'), "Skipping integration test.")
class TestVirtualGraphComposite(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        try:
            cls.qpu = DWaveSampler(solver=dict(flux_biases=True))
        except (ValueError, ConfigFileError, SolverNotFoundError):
            raise unittest.SkipTest("no qpu available")

    @classmethod
    def tearDownClass(cls):
        cls.qpu.client.close()

    def test_construction(self):
        child_sampler = self.qpu

        # get an embedding
        K10_edges = list(itertools.combinations(range(10), 2))
        embedding = minorminer.find_embedding(K10_edges, child_sampler.edgelist)

        sampler = VirtualGraphComposite(child_sampler, embedding,
                                        flux_bias_num_reads=10)

        dtest.assert_sampler_api(sampler)

        h = {}
        J = {edge: -1 for edge in K10_edges}

        # run with fbo
        sampler.sample_ising(h, J).resolve()

        # and again without
        sampler.sample_ising(h, J, apply_flux_bias_offsets=False).resolve()



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
# ================================================================================================
import os

from setuptools import setup


# change directories so this works when called from other locations. Useful in build systems.
setup_folder_loc = os.path.dirname(os.path.abspath(__file__))
os.chdir(setup_folder_loc)

exec(open(os.path.join(".", "dwave", "system", "package_info.py")).read())


install_requires = ['dimod>=0.10.0,<0.13.0',
                    'dwave-cloud-client>=0.9.1,<0.11.0',
                    'dwave-networkx>=0.8.10',
                    'dwave-preprocessing>=0.3',
                    'networkx>=2.0,<3.0',
                    'homebase>=1.0.0,<2.0.0',
                    'minorminer>=0.2.8,<0.3.0',
                    'numpy>=1.17.3,<2.0.0',
                    'dwave-greedy>=0.2.0',
                    'scipy>=1.5.2,<2.0.0',
                    ]

# NOTE: dwave-drivers can also be installed with `dwave install drivers`,
# and the exact requirements for `drivers` contrib package are defined in
# `dwave.system.package_info`, the `contrib` dict.
extras_require = {'drivers': ['dwave-drivers>=0.4.0,<0.5.0'],
                  }

python_requires = '>=3.7'

packages = ['dwave',
            'dwave.embedding',
            'dwave.system',
            'dwave.system.cache',
            'dwave.system.composites',
            'dwave.system.samplers',
            ]

setup(
    name='dwave-system',
    version=__version__,
    author=__author__,
    author_email=__authoremail__,
    description=__description__,
    long_description=open('README.rst').read(),
    url='https://github.com/dwavesystems/dwave-system',
    license='Apache 2.0',
    packages=packages,
    entry_points={
        'dwave_contrib': [
            'dwave-system = dwave.system.package_info:contrib'
        ]
    },
    install_requires=install_requires,
    extras_require=extras_require,
    python_requires=python_requires,
    zip_safe=False
)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest
import random

import dimod
import dwave_networkx as dnx

from dwave.system.testing import MockDWaveSampler
from dwave.system.composites import TilingComposite


class TestTiling(unittest.TestCase):
    def test_pegasus_single_cell(self):
        #Test trivial case of single cell (K4,4+4*odd) embedding over defect free
        mock_sampler = MockDWaveSampler(topology_type='pegasus')  # P3 structured sampler
        self.assertTrue('topology' in mock_sampler.properties and 'type' in mock_sampler.properties['topology'])
        self.assertTrue(mock_sampler.properties['topology']['type'] == 'pegasus' and 'shape' in mock_sampler.properties['topology'])
        sampler = TilingComposite(mock_sampler, 1, 1)
        h = {node: random.uniform(-1, 1) for node in sampler.structure.nodelist}
        J = {(u, v): random.uniform(-1, 1) for u, v in sampler.structure.edgelist}

        m = n = mock_sampler.properties['topology']['shape'][0] - 1
        expected_number_of_cells = m*n*3
        num_reads = 10
        response = sampler.sample_ising(h, J, num_reads=num_reads)
        self.assertTrue(sum(response.record.num_occurrences)==expected_number_of_cells*num_reads)
        
    def test_pegasus_multi_cell(self):
        #Test case of 2x3 cell embedding over defect free
        mock_sampler = MockDWaveSampler(topology_type='pegasus',topology_shape=[8])  # P8 structured sampler
        self.assertTrue('topology' in mock_sampler.properties and 'type' in mock_sampler.properties['topology'])
        self.assertTrue(mock_sampler.properties['topology']['type'] == 'pegasus' and 'shape' in mock_sampler.properties['topology'])
        sampler = TilingComposite(mock_sampler, 1, 1)
        h = {node: random.uniform(-1, 1) for node in sampler.structure.nodelist}
        J = {(u, v): random.uniform(-1, 1) for u, v in sampler.structure.edgelist}
        
        m_sub = 2
        n_sub = 3
        sampler = TilingComposite(mock_sampler, m_sub, n_sub)
        h = {node: random.uniform(-1, 1) for node in sampler.structure.nodelist}
        J = {(u, v): random.uniform(-1, 1) for u, v in sampler.structure.edgelist}

        m = n = mock_sampler.properties['topology']['shape'][0] - 1
        expected_number_of_cells = (m//m_sub)*(n//3)*3
        num_reads = 1
        response = sampler.sample_ising(h, J, num_reads = num_reads)
        self.assertTrue(sum(response.record.num_occurrences)==expected_number_of_cells*num_reads)
        
        
    def test_sample_ising(self):
        mock_sampler = MockDWaveSampler()  # C4 structured sampler

        sampler = TilingComposite(mock_sampler, 2, 2)

        h = {node: random.uniform(-1, 1) for node in sampler.structure.nodelist}
        J = {(u, v): random.uniform(-1, 1) for u, v in sampler.structure.edgelist}

        response = sampler.sample_ising(h, J)

    def test_tile_around_hole(self):

        # Create a Chimera C4 structured sampler with a node missing, so that
        # we have a defect pattern:
        # OOOX
        # OOOO
        # OOOO
        # OOOO
        # where O: complete cell, X: incomplete cell
        mock_sampler = MockDWaveSampler(broken_nodes=[8 * 3]) 
        hardware_graph = dnx.chimera_graph(4)  # C4

        # Tile with 2x2 cells:
        sampler = TilingComposite(mock_sampler, 2, 2, 4)
        # Given the above chimera graph, check that the embeddings are as
        # follows:
        # 00XX
        # 0011
        # 2211
        # 22XX
        # where 0,1,2: belongs to correspoding embedding, X: not used in any
        # embedding
        self.assertSetEqual({v for s in sampler.embeddings[0].values() for v in s},
                            {linear_index for linear_index, (i, j, u, k)
                             in hardware_graph.nodes(data='chimera_index')
                             if i in (0, 1) and j in (0, 1)})
        self.assertSetEqual({v for s in sampler.embeddings[1].values() for v in s},
                            {linear_index for linear_index, (i, j, u, k)
                             in hardware_graph.nodes(data='chimera_index')
                             if i in (1, 2) and j in (2, 3)})
        self.assertSetEqual({v for s in sampler.embeddings[2].values() for v in s},
                            {linear_index for linear_index, (i, j, u, k)
                             in hardware_graph.nodes(data='chimera_index')
                             if i in (2, 3) and j in (0, 1)})
        
    def test_tile_around_node_defects_pegasus(self):
        pegasus_shape = [5]
        # Create a pegasus P5 structured solver subject to node defects with the
        # following (nice-coordinate) 3x4x4 cell-level structure:
        # OOOX OOOO OOOO
        # OOOO OOOO OOOO
        # OOOO OOOO OOOO
        # OOOO OOOO OOOX
        # where O: complete cell, X: incomplete cell
        broken_node_nice_coordinates = [(0,0,3,0,1), (2,3,3,1,3)]
        broken_node_linear_coordinates = [
            dnx.pegasus_coordinates(pegasus_shape[0]).nice_to_linear(coord)
            for coord in broken_node_nice_coordinates]
        mock_sampler = MockDWaveSampler(topology_type='pegasus',
                                        topology_shape=pegasus_shape,
                                        broken_nodes=broken_node_linear_coordinates) 
        # Tile with 2x2 cells:
        sampler = TilingComposite(mock_sampler, 2, 2, 4)
        
        # Given the above pegasus graph, check that the embeddings are as
        # follows:
        # 00XX  3344 7788 
        # 0011  3344 7788 
        # 2211  5566 99XX  
        # 22XX  5566 99XX
        
        # Check correct number of embeddings and size of each is sufficient,
        # given chimera test checks detailed position:
        self.assertTrue(len(sampler.embeddings) == 10)
        self.assertFalse(any([len(emb) != 32 for emb in sampler.embeddings]))

        #Can be refined to check exact positioning, but a lot of ugly code:
        #For visualization in coordinate scheme use:
        #for emb in sampler.embeddings:
        #    print({key: dnx.pegasus_coordinates(pegasus_shape[0]).linear_to_nice(next(iter(val)))
        #           for key,val in emb.items()})

    def test_tile_around_edge_defects_pegasus(self):
        pegasus_shape = [5]

        # P5 structured sampler with one missing external edge that does not p
        # prevent tesselation of 2x2 blocks (12 tiles, equivalent to full yield)
        broken_edges_nice_coordinates = [(0,1,0,0,0), (0,2,0,0,0)]
        broken_edges = [tuple(
            dnx.pegasus_coordinates(pegasus_shape[0]).nice_to_linear(coord)
            for coord in broken_edges_nice_coordinates)]
        mock_sampler = MockDWaveSampler(topology_type='pegasus',
                                        topology_shape=pegasus_shape,
                                        broken_edges=broken_edges)
        sampler = TilingComposite(mock_sampler, 2, 2, 4)
        self.assertTrue(len(sampler.embeddings) == 12)
        
        # P5 structured sampler with one missing internal edge that prevents
        # tesselation of 2x2 blocks (otherwise 12 tiles, with edge defect 11)
        broken_edge_nice_coordinates = [(0,0,0,0,0), (0,0,0,1,0)]
        broken_edges = [tuple(
            dnx.pegasus_coordinates(pegasus_shape[0]).nice_to_linear(coord)
            for coord in broken_edge_nice_coordinates)]
        mock_sampler = MockDWaveSampler(topology_type='pegasus',
                                        topology_shape=pegasus_shape,
                                        broken_edges=broken_edges)
        sampler = TilingComposite(mock_sampler, 2, 2, 4)
        self.assertTrue(len(sampler.embeddings) == 11)
        
    def test_sample_ising(self):
        sampler = TilingComposite(MockDWaveSampler(), 2, 2)

        h = {node: random.uniform(-1, 1) for node in sampler.structure.nodelist}
        J = {(u, v): random.uniform(-1, 1) for u, v in sampler.structure.edgelist}

        response = sampler.sample_ising(h, J)

        # nothing failed and we got at least one response back per tile
        self.assertGreaterEqual(len(response), len(sampler.embeddings))

        for sample in response.samples():
            for v in h:
                self.assertIn(v, sample)

        for sample, energy in response.data(['sample', 'energy']):
            self.assertAlmostEqual(dimod.ising_energy(sample, h, J), energy)

    def test_sample_qubo(self):
        sampler = TilingComposite(MockDWaveSampler(), 2, 2)

        Q = {(u, v): random.uniform(-1, 1) for u, v in sampler.structure.edgelist}
        Q.update({(node, node): random.uniform(-1, 1) for node in sampler.structure.nodelist})

        response = sampler.sample_qubo(Q)

        # nothing failed and we got at least one response back per tile
        self.assertGreaterEqual(len(response), len(sampler.embeddings))

        for sample in response.samples():
            for u, v in Q:
                self.assertIn(v, sample)
                self.assertIn(u, sample)

        for sample, energy in response.data(['sample', 'energy']):
            self.assertAlmostEqual(dimod.qubo_energy(sample, Q), energy)

    def test_too_many_nodes(self):
        mock_sampler = MockDWaveSampler()  # C4 structured sampler

        sampler = TilingComposite(mock_sampler, 2, 2)

        h = {0: -1, 1: 1}
        J = {}

        response = sampler.sample_ising(h, J)

        __, num_columns = response.record.sample.shape

        self.assertEqual(num_columns, 2)
